# -*- coding: utf-8 -*-
"""RAWAPItry

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1_BbR5MsqLVTNUNaIzJrZUXnZEww0mG9R
"""

from __future__ import absolute_import, division, print_function, unicode_literals

"""
Created on June 11, 2019

@author: Jesse B. Hopkins

##############################################################################
 This file is part of RAW.

RAW is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RAW is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RAW.  If not, see <http://www.gnu.org/licenses/>.

##############################################################################

The purpose of this module is to provide an API for calling RAW functions from
other python programs. This is to provide easy access to RAW's functionality
in any data processing program you want to write in python
"""

from builtins import object, range, map, zip
# from io import open

# import os
import copy

# import tempfile
import traceback
import copy
import threading
import logging

import numpy as np
import pandas as pd
from scipy.optimize import curve_fit
from scipy import stats
from sklearn.metrics import r2_score

# raw_path = os.path.abspath(os.path.join('.', __file__, '..', '..'))
# if raw_path not in os.sys.path:
#     os.sys.path.append(raw_path)


from . import sascalc as SASCalc  # import bioxtasraw.sascalc as SASCalc
from . import (
    sasexceptions as SASExceptions,
)  # import bioxtasraw.SASExceptions as SASExceptions

# import bioxtasraw.SASFileIO as SASFileIO
# import bioxtasraw.SASMask as SASMask
from . import sasm as SASM  # import bioxtasraw.SASM as SASM

# import bioxtasraw.SASProc as SASProc
# import bioxtasraw.RAWSettings as RAWSettings
# import bioxtasraw.RAWGlobals as RAWGlobals
# import bioxtasraw.SECM as SECM
from . import bift as BIFT  # import bioxtasraw.BIFT as BIFT
# import bioxtasraw.DENSS as DENSS
# import bioxtasraw.SASUtils as SASUtils
# import bioxtasraw.RAWReport as RAWReport


def make_profile(q, i, err, name, q_err=None):
    """
    Makes a profile (:class:`bioxtasraw.SASM.SASM`) from q, I, and uncertainty
    vectors. All three input vectors must be the same length.

    Parameters
    ----------
    q: iterable
        The q vector for the scattering profile. Should be an iterable that can
        be cast to a :class:`numpy.array`, such as a list or :class:`numpy.array`.
    i: iterable
        The intensity vector for the scattering profile. Should be an iterable
        that can be cast to a :class:`numpy.array`, such as a list or
        :class:`numpy.array`.
    err: iterable
        The uncertainty vector for the scattering profile. Should be an iterable
        that can be cast to a :class:`numpy.array`, such as a list or
        :class:`numpy.array`.
    name: str
        The name of the profile. When loading a profile from a file, this is
        by default set as the filename (without the full path).
    q_err: iterable, optional
        The uncertainty vector in q for the scattering profile. Should be
        either None or an iterable that can be cast to a :class:`numpy.array',
        such as a list or :class:`numpy.array`. Default is None. Typically only
        used for SANS data.

    Returns
    -------
    profile: :class:`bioxtasraw.SASM.SASM`
        A scattering profile object that can be used with the other functions in
        the API.
    """

    profile = SASM.SASM(i, q, err, {"filename": name}, q_err)

    return profile


def bift(
    profile,
    idx_min=None,
    idx_max=None,
    pr_pts=100,
    alpha_min=150,
    alpha_max=1e10,
    alpha_pts=16,
    dmax_min=10,
    dmax_max=400,
    dmax_pts=10,
    mc_runs=300,
    use_guinier_start=True,
    single_proc=True,
    nprocs=None,
    settings=None,
):
    #### Maybe increase Dmax max and also maybe the minimum

    # def bift(profile, idx_min=None, idx_max=None, pr_pts=100, alpha_min=150,
    # alpha_max=1e10, alpha_pts=16, dmax_min=10, dmax_max=400, dmax_pts=10,
    # mc_runs=300, use_guinier_start=True, single_proc=True, nprocs=None,
    # settings=None):
    """
    Calculates the Bayesian indirect Fourier transform (BIFT) of a scattering
    profile to generate a P(r) function and determine the maximum dimension
    Dmax. Returns None and -1 values if BIFT fails.

    Parameters
    ----------
    profile: :class:`bioxtasraw.SASM.SASM`
        The profile to calculate the BIFT for.
    idx_min: int, optional
        The index of the q vector that corresponds to the minimum q point
        to be used in the IFT. Default is to use the first point of the q
        vector, unless use_guinier_start is set.
    idx_max: int, optional
        The index of the q vector that corresponds to the maximum q point
        to be used in the IFT. Default is to use the last point of the
        q vector.
    pr_pts: int, optional
        The number of points in the calculated P(r) function. This should
        be less than the number of points in the scattering profile.
        If settings are provided, this is overridden by the value in the
        settings.
    alpha_min: float, optional
        The minimum value of alpha for the parameter search step. If settings
        are provided, this is overridden by the value in the settings. The
        value of alpha can go beyond this bound in the optimization step,
        so this is not a hard limit on alpha.
    alpha_max: float, optional
        The maximum value of alpha for the parameter search step. If settings
        are provided, this is overridden by the value in the settings. The
        value of alpha can go beyond this bound in the optimization step,
        so this is not a hard limit on alpha.
    alpha_pts: int, optional
        The number of points in the alpha search space, which will be
        logarithmically spaced between alpha_min and alpha_max. If settings
        are provided, this is overridden by the value in the settings.
    dmax_min: float, optional
        The minimum value of Dmax for the parameter search step. If settings
        are provided, this is overridden by the value in the settings. The
        value of Dmax can go beyond this bound in the optimization step,
        so this is not a hard limit on Dmax.
    dmax_max: float, optional
        The maximum value of Dmax for the parameter search step. If settings
        are provided, this is overridden by the value in the settings. The
        value of Dmax can go beyond this bound in the optimization step,
        so this is not a hard limit on Dmax.
    dmax_pts: int, optional
        The number of points in the Dmax search space, which will be linearly
        spaced between dmax_min and dmax_max. If settings are provided, this
        is overridden by the value in the settings.
    mc_runs: int, optional
        The number of monte carlo runs used to generate the uncertainty
        estimates for the P(r) function.
    use_guiner_start: bool, optional
        If set to True, and no idx_min idx_min is provided, if a Guinier fit has
        been done for the input profile, the start point of the Guinier fit is
        used as the start point for the IFT.
    single_proc: bool, optional
        Whether to use one or multiple processors. Defaults to True. In limited
        testing the single processor version has been found to be 2-3x faster
        than the multiprocessor version, but actual results may depend on
        the computer and the number of gird search points.
    nprocs: int, optional
        If specified, and single_proc is False, determines the number of processors
        to use for BIFT. Otherwise defaults to number of processors in the computer
        -1 (minimum 1).
    settings: :class:`bioxtasraw.RAWSettings.RAWSettings`, optional
        RAW settings containing relevant parameters. If provided, the
        pr_Pts, alpha_min, alpha_max, alpha_pts, dmax_min, dmax_max, dmax_pts,
        and mc_runs parameters will be overridden with the values in the
        settings. Default is None.

    Returns
    -------
    ift: :class:`bioxtasraw.SASM.IFTM`
        The IFT calculated by BIFT from the input profile.
    dmax: float
        The maximum dimension of the P(r) function found by BIFT.
    rg: float
        The real space radius of gyration (Rg) from the P(r) function.
    i0: float
        The real space scattering at zero angle (I(0)) from the P(r) function.
    dmax_err: float
        The uncertainty in the maximum dimension of the P(r) function found
        by BIFT.
    rg_err: float
        The uncertainty in the real space radius of gyration (Rg) from the P(r)
        function.
    i0_err: float
        The uncertainty in the real space scattering at zero angle (I(0)) from
        the P(r) function.
    chi_sq: float
        The chi squared value of the fit of the scattering profile calculated
        from the P(r) function to the input scattering profile.
    log_alpha: float
        Log base 10 of the alpha value for the IFT.
    log_alpha_err: float
        Log base 10 of the uncertainty in the alpha value for the IFT.
    evidence: float
        The Bayesian evidence of the IFT.
    evidence_err: float
        The uncertainty in the Bayesian evidence of the IFT.
    """

    if settings is not None:
        pr_pts = settings.get("PrPoints")
        alpha_min = settings.get("minAlpha")
        alpha_max = settings.get("maxAlpha")
        alpha_pts = settings.get("AlphaPoints")
        dmax_min = settings.get("maxDmax")
        dmax_max = settings.get("minDmax")
        dmax_pts = settings.get("DmaxPoints")
        mc_runs = settings.get("mcRuns")

    q = profile.getQ()
    i = profile.getI()
    err = profile.getErr()
    filename = profile.getParameter("filename")

    if idx_min is None and use_guinier_start:
        analysis_dict = profile.getParameter("analysis")
        if "guinier" in analysis_dict:
            guinier_dict = analysis_dict[
                "guinier"
            ]  # **has a dictionary having a dictionary as key/value analysis_dict['guinier'] =info_dict
            idx_min = max(0, int(guinier_dict["nStart"]) - profile.getQrange()[0])
        else:
            idx_min = 0

    elif idx_min is None:
        idx_min = 0

    if idx_max is not None:
        q = q[idx_min : idx_max + 1]
        i = i[idx_min : idx_max + 1]
        err = err[idx_min : idx_max + 1]
    else:
        q = q[idx_min:]
        i = i[idx_min:]
        err = err[idx_min:]

    if nprocs is None:
        nprocs = 0

    bift_settings = {
        "npts": pr_pts,
        "alpha_max": alpha_max,
        "alpha_min": alpha_min,
        "alpha_n": alpha_pts,
        "dmax_min": dmax_min,
        "dmax_max": dmax_max,
        "dmax_n": dmax_pts,
        "mc_runs": mc_runs,
        "single_proc": single_proc,
        "nprocs": nprocs,
    }

    ift = BIFT.doBift(q, i, err, filename, **bift_settings)

    if ift is not None:
        dmax = float(ift.getParameter("dmax"))
        dmax_err = float(ift.getParameter("dmaxer"))
        rg = float(ift.getParameter("rg"))
        rg_err = float(ift.getParameter("rger"))
        i0 = float(ift.getParameter("i0"))
        i0_err = float(ift.getParameter("i0er"))
        chi_sq = float(ift.getParameter("chisq"))
        log_alpha = float(ift.getParameter("alpha"))
        log_alpha_err = float(ift.getParameter("alpha_er"))
        evidence = float(ift.getParameter("evidence"))
        evidence_err = float(ift.getParameter("evidence_er"))
        qmin = q[0]  # uncommented these 2
        qmax = q[-1]
        ##################################### NEED TO ADD TO GL GUINIER For Maybe do possible more fits

        results_dict = {}
        results_dict["Dmax"] = str(dmax)
        results_dict["Dmax_Err"] = str(dmax_err)
        results_dict["Real_Space_Rg"] = str(rg)
        results_dict["Real_Space_Rg_Err"] = str(rg_err)
        results_dict["Real_Space_I0"] = str(i0)
        results_dict["Real_Space_I0_Err"] = str(i0_err)
        results_dict["ChiSquared"] = str(chi_sq)
        results_dict["LogAlpha"] = str(log_alpha)
        results_dict["LogAlpha_Err"] = str(log_alpha_err)
        results_dict["Evidence"] = str(evidence)
        results_dict["Evidence_Err"] = str(evidence_err)
        results_dict["qStart"] = str(qmin)
        results_dict["qEnd"] = str(qmax)

        analysis_dict = profile.getParameter("analysis")
        analysis_dict["BIFT"] = results_dict
        profile.setParameter("analysis", analysis_dict)

    else:
        dmax = -1
        dmax_err = -1
        rg = -1
        rg_err = -1
        i0 = -1
        i0_err = -1
        chi_sq = -1
        log_alpha = -1
        log_alpha_err = -1
        evidence = -1
        evidence_err = -1
        qmin = q[0]
        qmax = q[-1]

    return (
        ift,
        dmax,
        rg,
        i0,
        dmax_err,
        rg_err,
        i0_err,
        chi_sq,
        log_alpha,
        log_alpha_err,
        evidence,
        evidence_err,
    )


def auto_guinier(profile, error_weight=True, single_fit=True, settings=None):
    """
    Automatically calculates the Rg and I(0) values from the Guinier fit by
    determining the best range for the Guinier fit.

    Parameters
    ----------
    profile: :class:`bioxtasraw.SASM.SASM`
        The profile to perform the Guineir fit on.
    error_weight: bool, optional
        If True (default), then the Guinier fit is calculated in an error
        weighted fashion. If not, the Guinier fit is calculated without
        error weight. This is overridden by the value in the settings if
        a settings object is provided.
    single_fit: bool, optional
        If True (default), then after the correct range for the Guinier fit
        is found a traditional Guinier fit is performed using that range. If
        False, currently the same is true. In the future, if False then the Rg
        and I(0) values may be averages over some range of best Guinier fit
        intervals.
    settings: :class:`bioxtasraw.RAWSettings.RAWSettings`, optional
        RAW settings containing relevant parameters. If provided, the
        error_weight parameter will be overridden with the value in the
        settings. Default is None.

    Returns
    -------
    rg: float
        The Rg value of the fit.
    i0: float
        The I(0) value of the fit.
    rg_err: float
        The uncertainty in Rg. This is calculated as the largest
        of the uncertainty returned from autorg and the uncertatiny as
        calculated from the covariance of the Guinier fit with the autorg
        determined ranges.
    i0_err: float
        The uncertainty in I(0). This is calculated as the largest
        of the uncertainty returned from autorg and the uncertatiny as
        calculated from the covariance of the Guinier fit with the autorg
        determined ranges.
    qmin: float
        The minimum q value of the Guinier fit.
    qmax: float
        The maximum q value of the Guinier fit.
    qRg_min: float
        The q*Rg value at the minimmum q value of the Guinier fit.
    qRg_max: float
        The q*Rg value at the maximum q value of the Guinier fit.
    idx_min: int
        The minimum index of the q vector used for Guinier fit.
    idx_max: int
        The maximum index of the q vector used for the GUinier fit.
    r_sqr: float
        The r^2 value of the fit.
    """

    if settings is not None:
        error_weight = settings.get("errorWeight")

    rg_auto, rger_auto, i0_auto, i0er_auto, idx_min, idx_max = SASCalc.autoRg(
        profile, single_fit, error_weight
    )

    if rg_auto != -1:
        q = profile.getQ()
        i = profile.getI()
        err = profile.getErr()

        rg_fit, i0_fit, rger_fit, i0er_fit, a, b = SASCalc.calcRg(
            q[idx_min : idx_max + 1],
            i[idx_min : idx_max + 1],
            err[idx_min : idx_max + 1],
            transform=True,
            error_weight=error_weight,
        )

        if single_fit:
            rg = float(rg_fit)
            i0 = float(i0_fit)
        else:
            rg = float(rg_auto)
            i0 = float(i0_auto)

        rg_err = max(float(rger_fit), float(rger_auto))
        i0_err = max(float(i0er_fit), float(i0er_auto))

        qmin = q[idx_min]
        qmax = q[idx_max]
        qRg_min = qmin * rg
        qRg_max = qmax * rg

        # Get fit r squared:
        x = np.square(q[idx_min : idx_max + 1])
        y = np.log(i[idx_min : idx_max + 1])
        y_fit = SASCalc.linear_func(x, a, b)
        error = y - y_fit
        r_sqr = 1 - np.square(error).sum() / np.square(y - y.mean()).sum()

        nmin_offset, _ = profile.getQrange()
        ##############ADDD TO G GUINIER #######################################################################
        info_dict = {}
        info_dict["Rg"] = rg
        info_dict["I0"] = i0
        info_dict["nStart"] = idx_min + nmin_offset
        info_dict["nEnd"] = idx_max + nmin_offset
        info_dict["qStart"] = qmin
        info_dict["qEnd"] = qmax
        info_dict["qRg_min"] = qRg_min
        info_dict["qRg_max"] = qRg_max
        info_dict["Rg_fit_err"] = rger_fit
        info_dict["I0_fit_err"] = i0er_fit
        info_dict["Rg_est_err"] = -1
        info_dict["I0_est_err"] = -1
        info_dict["Rg_autorg_err"] = rger_auto
        info_dict["I0_autorg_err"] = i0er_auto
        info_dict["Rg_err"] = rg_err
        info_dict["I0_err"] = i0_err
        info_dict["rsq"] = r_sqr

        analysis_dict = profile.getParameter("analysis")
        analysis_dict["guinier"] = info_dict
        profile.setParameter("analysis", analysis_dict)
    #############################################################################
    else:
        rg = -1.0
        i0 = -1.0
        rg_err = -1.0
        i0_err = -1.0
        qmin = -1.0
        qmax = -1.0
        qRg_min = -1.0
        qRg_max = -1.0
        idx_min = -1
        idx_max = -1
        r_sqr = -1.0
        # **I wonder why not return the dictionary (did they forget)- maybe that would make more effient ?
        # ** ANS don't need to bc they call the setParameter() method to object class SASM to add

    return (
        rg,
        i0,
        rg_err,
        i0_err,
        qmin,
        qmax,
        qRg_min,
        qRg_max,
        idx_min,
        idx_max,
        r_sqr,
    )


def Guiner_Data_Process(raw_data):
    """
    Gets the data in the format to be able to do the fit, like
    getting the ln(i) from a data frame made in GL guinier function call
    """
    q = np.array(raw_data["q"])
    i = np.array(raw_data["I"])
    q2 = np.power(q, 2)
    log_i = np.log(i)
    raw_data["q2"] = q2
    raw_data["ln(I)"] = log_i
    raw_data["II"] = i
    raw_data["q"] = q
    return raw_data


def Guiner_Fitting_V1(processed_data):
    line_fit = stats.linregress(xdata, ydata)  # does the fit
    slope = line_fit.slope  # finds slope
    yint.append(line_fit.intercept)
    r2 = line_fit.rvalue**2
    rg = (-slope * 3) ** (1 / 2)


def Guinier_GL(
    profile, settings=None
):  # P sure dont need settings here but just in cae
    """
    Runs Guinier method used at GL lab to match outputs with preferences from lab
    manual analysis, calculating Guinier fit by varying window size as well as using filtering
    to threshold solutions to be from the first low q points (qmin<0.005)
    """
    q = profile.getQ()
    i = profile.getI()
    err = profile.getErr()
    DF = pd.DataFrame()
    DF["q"] = q
    DF["I"] = i
    DF["Error"] = err
    # This function just performs math with q and I
    Test_Data = Guiner_Data_Process(DF)
    # filtering out the q2 regions above 0.005
    Test_Data = Test_Data.loc[DF["q2"] <= 0.005]  # Revise.
    ## Extending window fitting
    Rgs = []
    r2s = []
    Rg_max_Arr = []
    Rg_min_Arr = []  # added 0624
    N = []
    n_min = []
    n_max = []
    window_length = []
    i0_f = []
    qmin = []
    qmax = []
    # Variable window fitting. Will loop through each combination of starting index and window length for
    # each sample provided. A little time consuming, but not that bad for now.
    # Made window sligtly smaller as this is also done for BIOXTAS and polymer seem to have more an exponential curve so finding linear w large window hard
    for window in range(3, 40):
        for n in range(0, len(Test_Data) - window):
            q2 = Test_Data["q2"][n : n + window]
            q = Test_Data["q"][n : n + window]
            i_0 = Test_Data["II"][n : n + window]  # added 0624
            log_I = Test_Data["ln(I)"][n : n + window]
            res = stats.linregress(q2, log_I)
            slope = res.slope  # finds slope
            rg = (-slope * 3) ** (1 / 2)
            rg_max = rg * (np.array(q2)[-1] ** 0.5)
            Rg_max_Arr.append(rg_max)
            rg_min = rg * (np.array(q2)[0] ** 0.5)
            Rg_min_Arr.append(rg_max)
            Rgs.append(rg / 10)
            r2s.append(res.rvalue**2)
            n_min.append(n)
            n_max.append(n + window - 1)
            qmin.append(q.iloc[0])
            qmax.append(q.iloc[window - 1])
            i0_f.append(i_0.iloc[0])
            window_length.append(window)
            # qmin = q[0]
            # qmax = q[-1]
            # Storing all calculated values in a dictionary, enables easy conversion to a DataFrame

            # qmin = q.iloc[0] #don't need to add else bc defined here and same name as biox
            # qmax = q.iloc[window-1]
    dic = {
        "Rg1": Rgs,
        "R2": r2s,
        "qRg_Max": Rg_max_Arr,
        "N_Min": n_min,
        "N_Max": n_max,
        "Window Length": window_length,
        "i0": i0_f,
        "Qmin": qmin,
        "Qmax": qmax,
        "qRg_Min": Rg_min_Arr,
    }
    df = pd.DataFrame(data=dic)
    print(df)
    # Various filters being applied to the dataframe to downselect our fits.
    try:
        df = df.loc[df["N_Min"] <= 50]
        # df = df.loc[df['R2'] >= 0.90] #1
        print("max R2 ", max(df["R2"]))
        print("minqrmin ", min(df["qRg_Min"]))
        print("min N min ", min(df["N_Min"]))
        print("first filt", df)
        df = df.loc[df["qRg_Min"] <= 1.2]  # Changed Rgmax
        print("2nd filt", df)
        print("max R2 ", max(df["R2"]))
        print("minqrmin ", min(df["qRg_Min"]))
        print("min N min ", min(df["N_Min"]))
        df = df.loc[df["qRg_Max"] <= 1.8]
        print("3rd filt", df)
        print("max R2 ", max(df["R2"]))
        print("minqrmin ", min(df["qRg_Min"]))
        print("min N min ", min(df["N_Min"]))
        df = df.loc[df["R2"] >= 0.70]
        # df = df.loc[df['N_Min'] <= 50] #changed from 16
        df = df.sort_values(by=["R2"], ascending=False)[0:20].reset_index(drop=True)
        print(df)
        print(len(df))
        print(df.iloc[0])
        d_final = df.iloc[0]
    except IndexError:
        df = df.loc[df["N_Min"] <= 50]
        # df = df.loc[df['R2'] >= 0.90] #1
        print("max R2 ", max(df["R2"]))
        print("minqrmin ", min(df["qRg_Min"]))
        print("min N min ", min(df["N_Min"]))
        print("first filt", df)
        df = df.loc[df["qRg_Min"] <= 1.2]  # Changed Rgmax
        print("2nd filt", df)
        print("max R2 ", max(df["R2"]))
        print("minqrmin ", min(df["qRg_Min"]))
        print("min N min ", min(df["N_Min"]))
        df = df.loc[df["qRg_Max"] <= 1.8]
        print("3rd filt", df)
        print("max R2 ", max(df["R2"]))
        print("minqrmin ", min(df["qRg_Min"]))
        print("min N min ", min(df["N_Min"]))
        # df = df.loc[df['R2'] >= 0.70]
        # df = df.loc[df['N_Min'] <= 50] #changed from 16
        df = df.sort_values(by=["R2"], ascending=False)[0:20].reset_index(drop=True)
        print(df)
        print(len(df))
        print(df.iloc[0])
        d_final = df.iloc[0]
        # df = df.loc[df['R2'] >= 0.70]
        # print('first filt',df)
        # df = df.loc[df['qRg_Min'] <= 1.2] #Changed Rgmax
        # print('2nd filt',df)
        # df = df.loc[df['qRg_Max'] <= 1.8]
        # print('3rd filt',df)
        # df = df.loc[df['N_Min'] <= 50] #changed from 16
        # df = df.sort_values(by=['R2'],ascending=False)[0:20].reset_index(drop=True)
        # print(df)
        # print(len(df))
        # print(df.iloc[0])
        # d_final = df.iloc[0]

    d_final = df.iloc[0]
    # Ratchet but just making all the variables agree w bioxtas
    # rg = d_final['Rg1'][0]
    # i0 = d_final['i0'][0]
    # srt = d_final['N_Min'][0]
    # n_end= d_final['N_Max'][0]
    # qRg_max = d_final['qRg_Max'][0]
    # qRg_min = rg*qmin
    # r_sqr = d_final['R2'][0]
    # qmin = d_final['Qmin'][0]
    # qmax =d_final['Qmax'][0]

    rg = d_final["Rg1"]
    i0 = d_final["i0"]
    idx_min = d_final["N_Min"]
    idx_max = d_final[
        "N_Max"
    ]  # used to be called n_end now idx_max to mathc output form of BioXtas
    qRg_max = d_final["qRg_Max"]
    qRg_min = rg * d_final["Qmin"]
    r_sqr = d_final["R2"]
    qmin = d_final["Qmin"]
    qmax = d_final["Qmax"]
    # Window Length':window_length}
    # dic = {'Rg1':Rgs,'R2':r2s,'qRg_Max':Rg_max_Arr,'N_Min':n_min,'N_Max':n_max,'Window Length':window_length}
    info_dict = {}
    info_dict["Rg"] = rg  #
    info_dict["I0"] = i0  #
    info_dict["nStart"] = idx_min  # idx_min + nmin_offset
    info_dict["nEnd"] = idx_max  # idx_max + nmin_offset
    info_dict["qStart"] = qmin  #
    info_dict["qEnd"] = qmax  #
    info_dict["qRg_min"] = qRg_min  #
    info_dict["qRg_max"] = qRg_max  #
    info_dict["rsq"] = r_sqr
    # MIght wanna add these seem like just fit errs can get from fn
    info_dict["Rg_fit_err"] = None
    info_dict["I0_fit_err"] = None
    info_dict["Rg_est_err"] = None
    info_dict["I0_est_err"] = None
    info_dict["Rg_autorg_err"] = None
    info_dict["I0_autorg_err"] = None
    info_dict["Rg_err"] = None
    info_dict["I0_err"] = None
    analysis_dict = profile.getParameter("analysis")
    analysis_dict["guinier"] = info_dict
    profile.setParameter("analysis", analysis_dict)
    # og return from guinier fit bioxtas if errors maybe need to add those
    return (rg, i0, qmin, qmax, qRg_min, qRg_max, idx_min, idx_max, r_sqr)


def guinier_fit(profile, idx_min, idx_max, error_weight=True, settings=None):
    # This is the second function that can be used to do the fit that u been using before other than auto#

    """
    Calculates the Rg and I(0) values from the Guinier fit defined by the
    input idx_min and idx_max parameters.

    Parameters
    ----------
    profile: :class:`bioxtasraw.SASM.SASM`
        The profile to perform the Guineir fit on.
    idx_min: int
        The index of the q vector that corresponds to the minimum q point
        to be used in the Guinier fit.
    idx_max: int
        The index of the q vector that corresponds to the maximum q point
        to be used in the Guinier fit.
    error_weight: bool, optional
        If True (default), then the Guinier fit is calculated in an error
        weighted fashion. If not, the Guinier fit is calculated without
        error weight. This is overridden by the value in the settings if
        a settings object is provided.
    settings: :class:`bioxtasraw.RAWSettings.RAWSettings`, optional
        RAW settings containing relevant parameters. If provided, the
        error_weight parameter will be overridden with the value in the
        settings. Default is None.

    Returns
    -------
    rg: float
        The Rg value of the fit.
    i0: float
        The I(0) value of the fit.
    rg_err: float
        The uncertainty in Rg. This is calculated as the largest
        of the uncertainty returned from autorg and the uncertainty as
        calculated from the covariance of the Guinier fit with the autorg
        determined ranges.
    i0_err: float
        The uncertainty in I(0). This is calculated as the largest
        of the uncertainty returned from autorg and the uncertainty as
        calculated from the covariance of the Guinier fit with the autorg
        determined ranges.
    qmin: float
        The minimum q value of the Guinier fit.
    qmax: float
        The maximum q value of the Guinier fit.
    qRg_min: float
        The q*Rg value at the minimmum q value of the Guinier fit.
    qRg_max: float
        The q*Rg value at the maximum q value of the Guinier fit.
    r_sqr: float
        The r^2 value of the fit.
    """

    if settings is not None:
        error_weight = settings.get("errorWeight")

    q = profile.getQ()[idx_min : idx_max + 1]
    i = profile.getI()[idx_min : idx_max + 1]
    ierr = profile.getErr()[idx_min : idx_max + 1]

    # Remove NaN and Inf values:
    q = q[np.where(np.isfinite(i))]
    ierr = ierr[np.where(np.isfinite(i))]
    i = i[np.where(np.isfinite(i))]

    x = np.square(q)
    yerr = np.absolute(ierr / i)
    y = np.log(i)

    rg, i0, rger_fit, i0er_fit, a, b = SASCalc.calcRg(
        x, y, yerr, transform=False, error_weight=error_weight
    )

    rger_est, i0er_est = SASCalc.estimate_guinier_error(
        x, y, yerr, transform=False, error_weight=error_weight
    )

    if rger_est is None:
        rg_err = float(rger_fit)
    else:
        rg_err = max(float(rger_fit), float(rger_est))

    if i0er_est is None:
        i0_err = float(i0er_fit)
    else:
        i0_err = max(float(i0er_fit), float(i0er_est))

    # Get fit statistics:
    y_fit = SASCalc.linear_func(x, a, b)
    error = y - y_fit
    r_sqr = 1 - np.square(error).sum() / np.square(y - y.mean()).sum()

    qmin = profile.getQ()[idx_min]
    qmax = profile.getQ()[idx_max]
    qRg_min = qmin * rg
    qRg_max = qmax * rg

    info_dict = {}
    info_dict["Rg"] = rg
    info_dict["I0"] = i0
    info_dict["nStart"] = idx_min
    info_dict["nEnd"] = idx_max
    info_dict["qStart"] = qmin
    info_dict["qEnd"] = qmax
    info_dict["qRg_min"] = qRg_min
    info_dict["qRg_max"] = qRg_max
    info_dict["Rg_fit_err"] = rger_fit
    info_dict["I0_fit_err"] = i0er_fit
    info_dict["Rg_est_err"] = rger_est
    info_dict["I0_est_err"] = i0er_est
    info_dict["Rg_autorg_err"] = -1
    info_dict["I0_autorg_err"] = -1
    info_dict["Rg_err"] = rg_err
    info_dict["I0_err"] = i0_err
    info_dict["rsq"] = r_sqr

    analysis_dict = profile.getParameter("analysis")
    analysis_dict["guinier"] = info_dict
    profile.setParameter("analysis", analysis_dict)

    return rg, i0, rg_err, i0_err, qmin, qmax, qRg_min, qRg_max, r_sqr



from setuptools import setup, find_packages

setup(
    name="saxs_assistant",
    version="0.1.1.dev7",
    description="SAXS Assistant: Automated analysis of SAXS data including Guinier, PDDF, and ML-based Dmax prediction",
    author="Cesar Ramirez",
    author_email="cr828@scarletmai.rutgers.edu",  # Update this
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    package_data={"saxs_assistant": ["models/*.joblib"]},
    install_requires=[
        "pandas==2.2.2",
        "joblib==1.5.1",
        "matplotlib==3.10.3",
        "numba==0.61.2",
        "numpy==1.26.4",
        "scikit-learn==1.6.1",
        "scipy==1.15.3",
        "tqdm==4.67.1",
        "openpyxl==3.1.5",
        "pillow==11.2.1",
        "requests==2.32.4",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",  # Or your chosen license
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)



"""This is the main runner for SAXS Assistant. Broken into multiple try blocks to prevent
overall failure if certain analysis cannot be done.
It runs the SAXS analysis on a DataFrame of that has the name of the files to analyze.
Extracting various features and generating plots that are stored in a dictionary.
This module is designed to be run as a script, and it will process the SAXS data files specified in the input DataFrame.
"""

import numpy as np
import pandas as pd
import warnings
from IPython.display import display, Javascript
import time
from pathlib import Path
import os
from tqdm import tqdm
import logging
from joblib import load, dump
from .rawutils import sasm as SASM

from .rawutils import Profile_Loader
from .rawutils import rawapitry as raw
from .rg_tools import (
    select_final_rg_from_candidates,
    rg_method_1,
    select_best_rg_method,
)
from .utils.helpers import reprocess_sasbdb_q_values
from .features import get_GPA, get_kratky
from .ML import (
    predict_dmax_from_features_only,
    assign_gmm_clusters,
    compute_franke_features,
)


from .PDDF_tools import (
    get_all_pr_results,
    unpack_pr_fits_dict,
    select_best_pr_fit,
)

# import select_best_pr_fit
from .rawutils import sasexceptions as SASEceptions

# Setup logger
logging.basicConfig(level=logging.INFO)

# Helper to store plots
plot_data = {}


def get_unique_output_dir(base_dir="results"):
    """Create a unique output directory by appending an index if the base directory already exists.
    Used in the analyze_and_save function to ensure results are saved in a unique directory, for when multiple instances of script are ran simultaneously.
    This prevents overwriting results from previous runs. Also usful for future steps when updating and combining results.
    :param base_dir: The base directory name to create.
    :return: A Path object pointing to the unique directory.
    """
    base_path = Path(base_dir)
    i = 1
    while base_path.exists():
        base_path = Path(f"{base_dir}_{i}")
        i += 1
    base_path.mkdir(parents=True)
    return base_path


def beep():
    """
    Used for notifications
    """
    display(
        Javascript("""
      (async () => {
          const context = new AudioContext();
          const o = context.createOscillator();
          const g = context.createGain();
          o.type = "sine";
          o.connect(g);
          g.connect(context.destination);
          o.start();
          g.gain.exponentialRampToValueAtTime(0.00001, context.currentTime + 1);
      })();
  """)
    )


def save_plot(sample_id, name, plot_dict):
    plot_data.setdefault(sample_id, {})[name] = plot_dict


# Helper to update a row in murthy_df
def update_murthy_df_row(df, index, updates):
    for key, value in updates.items():
        df.loc[df.index[index], key] = value


# Helper to check if Guinier fit is acceptable
def meets_guinier_criteria(r2, qrgmin, qrgmax):
    return r2 > 0.74 and qrgmin < 0.9 and qrgmax < 1.4


# Helper to calculate residuals
def calculate_residuals(i_fit, i_orig, err_orig):
    residuals = i_orig - i_fit
    normalized = residuals / err_orig
    return residuals, normalized, np.sum(normalized**2) / len(i_fit)


# Helper to get closest match to Guinier Rg


def get_best_pr_match(guinier_rg, rg_min):
    errors = [abs((guinier_rg - rg) / rg) * 100 for rg in rg_min]
    return int(np.argmin(errors)), min(errors)


# Helper to build fin_pr DataFrame
def build_fin_pr_df(
    indices, rg_min, i0_p_list, chi_sq_list, dmax_mins, logas, dmax_err
):
    data = []
    for i in indices:
        data.append(
            [
                i,
                dmax_mins[i],
                round(chi_sq_list[i], 3),
                round(logas[i], 2),
                round(dmax_err[i], 3),
                i0_p_list[i],
                rg_min[i],
            ]
        )
    return pd.DataFrame(
        data,
        columns=["Index", "Dmax", "Chi^2", "Log alpha", "Dmax Error", "Pr i0", "Pr Rg"],
    )


# Start and end time tracker
def track_time():
    start_time = time.time()
    print(
        f"Start Time: {time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(start_time))}"
    )
    return start_time


def print_end_time(start_time):
    end_time = time.time()
    print(f"End Time: {time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(end_time))}")
    print(f"Total Execution Time: {end_time - start_time:.2f} seconds")


# multiple try blocks-Prevents overall failure if certain analysis cannot be done
logging.basicConfig(level=logging.ERROR)


# multiple try blocks-Prevents overall failure if certain analysis cannot be done


def run_analysis(df_wrong, s0=0):
    start_time = track_time()
    df_wrong = df_wrong.reset_index(drop=True)

    for j in tqdm(range(s0, len(df_wrong)), desc="Running SAXS Analysis"):
        sample_id = df_wrong["file name"][j]
        logging.info(f"Analyzing {sample_id} ({j + 1}/{len(df_wrong)})")

        skip_sample = False  # Flagging to skip sample if any error occurs

        try:
            profile = Profile_Loader.load_profiles(
                df_wrong["path"][j] + df_wrong["file name"][j]
            )[0]
            # This resamples data to meet format as GL collection
            # Overall works well, tested with 3,000+ SASBDB files, this doesn't impact and allows faster execution
            # Serves to also use number of points in fit to evaluate fit, as if very low binning, even if have
            # multiple points, the Rg q interval is not good, this way can have a standard on window acceptance
            # if don't want to do this then can just comment out, and add the/10 for files in nm
            if df_wrong["Angular unit"][j] == "1/A":
                reprocessed_arrays = reprocess_sasbdb_q_values(
                    profile._q_raw, profile._i_raw, profile._err_raw
                )
            elif df_wrong["Angular unit"][j] == "1/nm":
                reprocessed_arrays = reprocess_sasbdb_q_values(
                    profile._q_raw / 10, profile._i_raw, profile._err_raw
                )
            profile._q_raw = reprocessed_arrays["q"]
            profile._i_raw = reprocessed_arrays["I"]
            profile._err_raw = reprocessed_arrays["error"]

            q_raw, I_raw, err_raw = profile._q_raw, profile._i_raw, profile._err_raw

            profile.q = profile._q_raw
            profile.i = profile._i_raw
            profile.err = profile._err_raw

            mask = profile.q < 0.255
            q, I, err = profile.q[mask], profile.i[mask], profile.err[mask]
            profile.q = q
            profile.i = I
            profile.err = err

            # Log Profile
            try:
                # used to be plot_2 now profile
                save_plot(
                    sample_id,
                    "profile",
                    {
                        "type": "errorbar",
                        "x": q,
                        "y": I,
                        "yerr": err,
                        "ecolor": "g",
                        "fmt": "o",
                        "yscale": "log",
                        "title": f"Profile of {sample_id}",
                        "xlabel": "q ($\\AA^{-1}$)",
                        "ylabel": "Log(I(q))",
                        "q_truncated": q,
                        "I_truncated": I,
                        "err_truncated": err,
                        "q_untruncated": profile._q_raw,  # these are the untruncated values but sampled
                        "I_untruncated": profile._i_raw,
                        "err_untruncated": profile._err_raw,
                    },
                )
                # plt.close(fig)
            except Exception as e:
                logging.warning(f"Log profile plot failed for {sample_id}: {e}")

        except Exception as e:
            logging.warning(f"Failed to load profile for {sample_id}: {e}")
            df_wrong.loc[df_wrong.index[j], "Fatal Error"] = "Profile Load"
            plot_data.setdefault(sample_id, {})["Flagged"] = True
            print(
                'Did you remember to setup the dataframe with the "path" and "file name" columns?'
            )
            print(
                "If not use the `prepare_dataframe(dataframe_path, data_folder_apth)` function to do so."
            )
            print("for more information type prepare_dataframe??")
            continue

        # GPA Plot
        try:  # plot_1 before now GPA
            gpa_data = get_GPA(I, q, plot=False)
            save_plot(
                sample_id,
                "GPA",
                {
                    "type": "scatter",
                    "x": gpa_data["x"],
                    "y": gpa_data["y"],
                    "title": "GPA: Guinier Region Existence Check",
                    "xlabel": "$q^2$",
                    "ylabel": "qI(q)",
                },
            )
        except Exception as e:
            logging.warning(f"GPA plot failed for {sample_id}: {e}")

        # Kratky
        try:
            kratky_data = get_kratky(q, I, plot=False)
            save_plot(
                sample_id,
                "kratky",
                {
                    "type": "line",
                    "x": kratky_data["x"],
                    "y": kratky_data["y"],
                    "xlabel": "q",
                    "ylabel": "$q^2 I(q)$",
                    "title": "Kratky Plot",
                },
            )

        except Exception as e:
            logging.warning(f"Kratky plot failed for {sample_id}: {e}")

        # AutoRg
        try:
            (
                rg_auto,
                i0_auto,
                rg_err,
                i0_err,
                qmin,
                qmax,
                qrg_min,
                qrg_max,
                idx_min,
                idx_max,
                r_sq,
            ) = raw.auto_guinier(profile)

            update_murthy_df_row(
                df_wrong,
                j,
                {
                    "AutoRg Rg": rg_auto,
                    "AutoRg I0": i0_auto,
                    "AutoRg Rg Err": rg_err,
                    "AutoRg I0 Err": i0_err,
                    "AutoRg qmin": qmin,
                    "AutoRg qmax": qmax,
                    "AutoRg qRg Min": qrg_min,
                    "AutoRg qRg Max": qrg_max,
                    "AutoRg Guinier R²": r_sq,
                    "AutoRg Rg idx min": idx_min,
                    "AutoRg Rg idx max": idx_max,
                },
            )
            plot_data.setdefault(sample_id, {})["Auto Rg"] = {
                "nmin": idx_min,
                "nmax": idx_max,
                "Rg": rg_auto,
                "i0": i0_auto,
                "Rg Err": rg_err,
                "I0 Err": i0_err,
                "qmin": qmin,
                "qmax": qmax,
                "fit qRg Min": qrg_min,
                "fit qRg Max": qrg_max,
                "R2": r_sq,
            }

        except Exception as e:
            logging.warning(f"AutoRg failed for {sample_id}: {e}")
            skip_sample = True  # Flag to skip further analysis for this sample
        if skip_sample:
            df_wrong.loc[df_wrong.index[j], "Fatal Error"] = "Auto Gunier"
            plot_data.setdefault(sample_id, {})["Flagged"] = True
            continue

        # P(r)
        try:
            pr_fits_dict = get_all_pr_results(profile, q, I, err)
            plot_data.setdefault(sample_id, {})["pr_fits"] = pr_fits_dict
            # sample_fits = plot_data["SASDPB9.dat"]["pr_fits"] # This is how can access the Prs for example

            update_murthy_df_row(
                df_wrong,
                j,
                {
                    "Num P(r) Fits": len(pr_fits_dict),
                    "Avg P(r) Rg": np.mean([v["Rg"] for v in pr_fits_dict.values()]),
                },
            )
        except Exception as e:
            logging.warning(f"P(r) fit failed for {sample_id}: {e}")

        # Here now running Method 1 For Rg
        try:
            pr_rg_list = [fit["Rg"] for fit in pr_fits_dict.values()]
            rg_method1_df, rg_method1_all = rg_method_1(q, I, err, pr_rg_list)

            if rg_method1_df is not None:
                plot_data.setdefault(sample_id, {})["Rg Method 1"] = {
                    "candidates_df": rg_method1_df.to_dict(orient="list"),
                    "raw_dicts": rg_method1_all,
                }

                update_murthy_df_row(
                    df_wrong,
                    j,
                    {
                        "Method 1 Rg Count": len(rg_method1_df),
                        "Method 1 Rg Mean": round(rg_method1_df["Rg"].mean(), 2),
                        "Method 1 Best R²": round(rg_method1_df["fit_r2"].max(), 3),
                    },
                )

            # best_fit = rg_method1_df.sort_values(by='fit_r2', ascending=False).iloc[0]
            # update_murthy_df_row(df_wrong, j, {
            #     'Method 1 Rg': best_fit['Rg'],
            #     'Method 1 I0': best_fit['i0'],
            #     'Method 1 qRgmin': best_fit['qRgmin'],
            #     'Method 1 qRgmax': best_fit['qRgmax'],
            #     'Method 1 R²': best_fit['fit_r2'],
            #     'Method 1 Residual Width': best_fit['Res window'],
            #     'Method 1 GPA Peak': best_fit['GPA x peak'],
            #     'Method 1 Peak/Xcross': best_fit['peaks/x-cross']
            # })

        except Exception as e:
            logging.warning(f"Rg Method 1 failed for {sample_id}: {e}")
            skip_sample = True  # Flag to skip further analysis for this sample
        if skip_sample:
            df_wrong.loc[df_wrong.index[j], "Fatal Error"] = "Method 1 Initialization"
            plot_data.setdefault(sample_id, {})["Flagged"] = True
            continue

        # Here doing filtering to get final Rg from method 1
        try:
            final_method1_df = select_final_rg_from_candidates(
                rg_method1_all, q, I, err, sample_id
            )

            if final_method1_df is not None:
                # best_fit = final_method1_df.sort_values(by='fit_r2', ascending=False).iloc[0]
                best_fit = final_method1_df.iloc[0]

                plot_data.setdefault(sample_id, {})["Rg Method 1 Final"] = (
                    final_method1_df.to_dict(orient="list")
                )
                update_murthy_df_row(
                    df_wrong,
                    j,
                    {
                        "Method 1 Final Rg": best_fit["Rg"],
                        "Method 1 Final Rg Err": best_fit["Rg Err"],  # NEW
                        "Method 1 Final I0": best_fit["i0"],  # New
                        "Method 1 Final I0 Err": best_fit["I0 Err"],  # NEW
                        "Method 1 Final R²": best_fit["fit_r2"],
                        "Method 1 qRgmin": best_fit["qRgmin"],
                        "Method 1 qRgmax": best_fit["qRgmax"],
                        "Method 1 GPA Peak": best_fit["GPA x peak"],
                        "Method 1 Res Width": best_fit["Res window"],
                        "Method 1 nmin": best_fit["nmin"],
                        "Method 1 nmax": best_fit["nmax"],
                        "Method 1 Peak/Xcross": best_fit["peaks/x-cross"],
                        "Method 1 GPA x peak": best_fit["GPA x peak"],
                        "Method 1 GPA y peak": best_fit["GPA y peak"],
                    },
                )

            else:
                logging.warning(f"Method 1 could not resolve Rg for {sample_id}")
                skip_sample = True  # Flag to skip further analysis for this sample
                if skip_sample:
                    df_wrong.loc[df_wrong.index[j], "Fatal Error"] = "Method 1 Step 2"
                    plot_data.setdefault(sample_id, {})["Flagged"] = True

                    continue
        except Exception as e:
            logging.warning(f"Rg Method 1 final selection failed for {sample_id}: {e}")
            skip_sample = True  # Flag to skip further analysis for this sample
            if skip_sample:
                df_wrong.loc[df_wrong.index[j], "Fatal Error"] = (
                    "Method 1 Final Selection"
                )
                plot_data.setdefault(sample_id, {})["Flagged"] = True
                continue

        # Here now making selection for Final Rg based on Mean residuals
        try:
            selection = select_best_rg_method(
                q,
                I,
                err,
                rg_auto,
                i0_auto,
                best_fit["Rg"],
                best_fit["i0"],
                rg_auto_err=df_wrong.loc[j, "AutoRg Rg Err"],
                i0_auto_err=df_wrong.loc[j, "AutoRg I0 Err"],
                rg1_err=best_fit["Rg Err"],
                i01_err=best_fit["I0 Err"],
                r2_auto=df_wrong.loc[j, "AutoRg Guinier R²"],
                r2_1=best_fit["fit_r2"],
                sample_id=sample_id,
            )

            if selection:
                update_murthy_df_row(
                    df_wrong,
                    j,
                    {
                        "Final Rg": selection["Final Rg"],
                        "Final qRg min": selection["Final qRg min"],
                        "Final qRg max": selection["Final qRg max"],
                        "Final I0": selection["Final I0"],
                        "Final Rg Err": selection["Final Rg Err"],
                        "Final I0 Err": selection["Final I0 Err"],
                        "Final Rg R²": selection["Final R²"],
                        "Rg Method": selection["Selected Method"],
                        "Final Rg Residual Mean": selection["Residual Stats"][
                            "residual_mean"
                        ][0],
                        "Final Rg Reduced Chi²": selection["Residual Stats"][
                            "Reduced Chi Squared"
                        ][0],
                    },
                )
                if selection["Selected Method"] == "Method 1":
                    df_wrong.loc[df_wrong.index[j], "Final G qRgmin"] = df_wrong.loc[
                        df_wrong.index[j], "Method 1 qRgmin"
                    ]
                    df_wrong.loc[df_wrong.index[j], "Final G qRgmax"] = df_wrong.loc[
                        df_wrong.index[j], "Method 1 qRgmax"
                    ]
                elif selection["Selected Method"] == "AutoRg":
                    df_wrong.loc[df_wrong.index[j], "Final G qRgmin"] = df_wrong.loc[
                        df_wrong.index[j], "AutoRg qRg Min"
                    ]
                    df_wrong.loc[df_wrong.index[j], "Final G qRgmax"] = df_wrong.loc[
                        df_wrong.index[j], "AutoRg qRg Max"
                    ]

                plot_data.setdefault(sample_id, {})["Rg Selection"] = selection
        except Exception as e:
            logging.warning(f"Final Rg selection failed for {sample_id}: {e}")
            skip_sample = True  # Flag to skip further analysis for this sample
            if skip_sample:
                df_wrong.loc[df_wrong.index[j], "Fatal Error"] = "Final Rg Selection"
                plot_data.setdefault(sample_id, {})["Flagged"] = True
                continue

        # Here selecting a Pr from the ones already did

        try:
            pr_fits_dict = plot_data[sample_id]["pr_fits"]
            (
                rg_min,
                pr_rg_err_list,
                i0_p_list,
                pr_i0_err_list,
                chi_sq_list,
                dmax_mins,
                logas,
                dmax_err,
                pr_qmin_list,
                pr_qmax_list,
                pr_list,
                pr_i_orig,
                pr_fit,
                pr_err_orig,
                pr_q_orig,
                pr_qxt,
                nmins,
            ) = unpack_pr_fits_dict(pr_fits_dict)

            fin_pr, best_idx = select_best_pr_fit(
                rg_min,
                pr_rg_err_list,
                i0_p_list,
                pr_i0_err_list,
                chi_sq_list,
                dmax_mins,
                logas,
                dmax_err,
                pr_qmin_list,
                pr_qmax_list,
                df_wrong["Final Rg"][j],
                df_wrong["Final I0"][j],
                pr_list,
                pr_i_orig,
                pr_fit,
                pr_err_orig,
                pr_q_orig,
                pr_qxt,
                nmins,
                q,
                I,
                err,
                sample_id,
                df_wrong,
                j,
            )

            # Plot 6: Final P(r) used to be called plot_6
            plot_data[sample_id]["pr_plot"] = {
                "type": "scatter",
                "x": pr_list[best_idx][0],
                "y": pr_list[best_idx][1],
                "Rg": rg_min[best_idx],
                "i0": i0_p_list[best_idx],
                "title": f"P(r) Rg: {rg_min[best_idx]}%\n"
                f"Dmax: {round(dmax_mins[best_idx], 0)} +/- {round(fin_pr['Dmax Error'][0], 0)}",
                "xlabel": "r",
                "ylabel": "P(r)",
                "Rgs considered": rg_min,
                "i0s considered": i0_p_list,
                "pr_i_orig": pr_i_orig[best_idx],
                "pr_err_orig": pr_err_orig[best_idx],
                "pr_q_orig": pr_q_orig[best_idx],
                "pr_fit": pr_fit[best_idx],
                "nmins": nmins[best_idx],
                "chis considered": chi_sq_list,
            }
            # Since added the errors
            plot_data[sample_id]["pr_plot"].update(
                {
                    "Rg err": pr_rg_err_list[best_idx],
                    "i0 err": pr_i0_err_list[best_idx],
                    "qmin": pr_qmin_list[best_idx],
                    "qmax": pr_qmax_list[best_idx],
                }
            )

            # Plot 7: Residuals
            residuals = pr_i_orig[best_idx] - pr_fit[best_idx]
            residuals_x = q[nmins[best_idx] :]
            residuals_y = residuals / pr_err_orig[best_idx]
            normalized_residuals = residuals_y
            # plot_7 now called pr_residuals
            plot_data[sample_id]["pr_residuals"] = {
                "type": "scatter",
                "q_extrapolated": pr_qxt[best_idx],
                "i_fit": pr_fit[best_idx],
                "x_line": q[nmins[best_idx] :],
                "y_line": I[nmins[best_idx] :],
                "title": f"Fit/Data $X^2$: {round(chi_sq_list[best_idx], 3)}%\n"
                f"Dmax +/- {round(fin_pr['Dmax Error'][0], 0)}",
                "xlabel": "q",
                "ylabel": "I(q)",
                "residuals_x": residuals_x,
                "residuals_y": residuals_y,
                "Chi_sqr": chi_sq_list[best_idx],
                "normalized residuals": normalized_residuals,
                "gi_bift_i_orig": pr_i_orig[best_idx],
                "err_orig": pr_err_orig[best_idx],
                "nmin": nmins[best_idx],
            }

        except Exception as e:
            logging.warning(
                f"Failed to extract and select final P(r) for {sample_id}: {e}"
            )
            skip_sample = True  # Flag to skip further analysis for this sample
            if skip_sample:
                df_wrong.loc[df_wrong.index[j], "Fatal Error"] = "Final P(r) Selection"
                plot_data.setdefault(sample_id, {})["Flagged"] = True
                continue

        # --- Franke Feature Extraction and Dimensionless Kratky Storage ---
        try:
            final_rg = df_wrong.loc[j, "Final Rg"]
            final_i0 = df_wrong.loc[j, "Final I0"]

            if (
                pd.notna(final_rg)
                and pd.notna(final_i0)
                and final_rg > 0
                and final_i0 > 0
            ):
                # Compute Franke features (no plotting)
                df_wrong = compute_franke_features(
                    q, I, final_rg, final_i0, sample_id, df_wrong, j
                )

                # Now optionally store Dimless Kratky in plot_data
                qRg = q_raw * final_rg
                IqRg2 = ((qRg) ** 2) * I_raw / final_i0

                plot_data.setdefault(sample_id, {})["Dimless Kratky"] = {
                    "type": "line",
                    "x": qRg,
                    "y": IqRg2,
                    "xlabel": "qRg",
                    "ylabel": "$(qRg)^2 I(q)/I(0)$",
                    "title": "Dimensionless Kratky",
                    "qRg": qRg,
                    "qRg^2": IqRg2,
                }

            else:
                logging.warning(
                    f"Skipping Franke features, & Dim Kratky for {sample_id} due to invalid Rg/I0"
                )
                skip_sample = True  # Flag to skip further analysis for this sample
                if skip_sample:
                    df_wrong.loc[df_wrong.index[j], "Fatal Error"] = "Franke Features"
                    continue
            # Here adding the clustering and probabilities
            try:
                gmm_results = assign_gmm_clusters(df_wrong, j)
                if gmm_results:
                    plot_data.setdefault(sample_id, {})["GMM Clustering"] = gmm_results
            except Exception as e:
                logging.warning(f"GMM block failed for {sample_id}: {e}")

        except Exception as e:
            logging.warning(f"Franke block failed for {sample_id}: {e}")
        try:
            pred_dmax, dmax_feats = predict_dmax_from_features_only(df_wrong, j)

            df_wrong.loc[df_wrong.index[j], "Predicted Dmax"] = pred_dmax

            if isinstance(dmax_feats, dict) and "error" not in dmax_feats:
                plot_data.setdefault(sample_id, {})["Predicted Dmax"] = {
                    "features_used": dmax_feats,
                    "prediction": pred_dmax,
                }
            else:
                plot_data.setdefault(sample_id, {})["Predicted Dmax"] = {
                    "error": dmax_feats.get("error", "Unknown error")
                }

        except Exception as e:
            logging.warning(f"Final Dmax prediction block failed for {sample_id}: {e}")

    print_end_time(start_time)
    return plot_data, df_wrong


def analyze_and_save(df_path, start_index=0, end_index=None, output_dir=None):
    """
    Main function to analyze SAXS data from a DataFrame and save results.
    Same as run_analysis but with input and output handling.

    :param df_path: Path to the input Excel file containing the file names, path and angular units.
    :param start_index: Index to start analysis from (default is 0). Useful for analyzing large sets in portions
    :param end_index: If specified, index to stop analysis at (exclusive). Default is None which means all data.
    :param output_dir: Directory to save results. Optional, if not specified, a 'return' folder will be created next to the input file.
    :return: Tuple of plot_data dictionary and updated DataFrame with results.
    """

    def sanitize_path(path_str):
        return Path(str(path_str).strip().strip('"').strip("'"))

    def clean_dataframe_paths(df, column_name):
        df[column_name] = df[column_name].apply(
            lambda x: str(Path(str(x).strip().strip('"').strip("'")))
        )
        return df

    # Clean up and parse input path
    df_path = sanitize_path(df_path)

    # If output_dir is not specified, create a 'return' folder next to the input file
    if output_dir is None:
        output_dir = df_path.parent / "return"
    else:
        output_dir = sanitize_path(output_dir)

    print(f"Using input file: {df_path}")
    print(f"Saving results to: {output_dir}")

    # Load and clean dataframe
    df = pd.read_excel(df_path)
    df = clean_dataframe_paths(df, "path")
    df["path"] = df["path"] + "/"

    # Run analysis
    if end_index is None:
        plot_data, updated_df = run_analysis(df, start_index)
    elif int(end_index) > len(df):
        raise ValueError(
            f"end_index {end_index} is greater than length of df {len(df)}"
        )
    else:
        plot_data, updated_df = run_analysis(df[: int(end_index)], start_index)

    # Save results
    # Only apply auto-folder logic if output_dir wasn't specified
    # This makes sure no overwriting happens if multiple instances of the script are run simultaneously
    if output_dir == "outputs":
        output_dir = get_unique_output_dir("results")
    else:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    dump(plot_data, output_dir / "plot_data.joblib")
    updated_df.to_excel(output_dir / "results.xlsx")
    beep()  # Notify user that analysis is complete
    return plot_data, updated_df


import pandas as pd
from pathlib import Path
from datetime import datetime


def prepare_dataframe(dataframe_path=None, folder_path=None, angular_unit=None):
    """
    Prepares and saves a dataframe for SAXS analysis with required columns:
    - 'file name'
    - 'path'
    - 'Angular unit'

    Parameters:
    ----------
    dataframe_path : str or Path, optional
        Path to an existing Excel file with 'file name' and 'path' columns.
    folder_path : str or Path, optional
        If dataframe_path is not provided, use this to create a new dataframe.
    angular_unit : str
        Unit of the angular axis. Can be '1/A', 'A', '1/nm', or 'nm'.

    Returns:
    -------
    pd.DataFrame
        Formatted dataframe ready for analysis.
    """
    print("folder_path before parsing:", folder_path)

    if angular_unit is None:
        raise ValueError(
            "You must provide the 'angular_unit' argument (e.g., '1/A', 'nm')."
        )

    # Normalize angular unit
    angular_unit = angular_unit.strip().lower()
    if angular_unit in ["a", "1/a"]:
        angular_unit = "1/A"
    elif angular_unit in ["nm", "1/nm"]:
        angular_unit = "1/nm"
    else:
        raise ValueError("angular_unit must be '1/A' or '1/nm'")

    save_path = None

    if dataframe_path:
        dataframe_path = Path(str(dataframe_path).strip().strip('"').strip("'"))
        df = pd.read_excel(dataframe_path)

        if "path" not in df.columns or "file name" not in df.columns:
            raise ValueError("Dataframe must contain 'file name' and 'path' columns.")

        # Clean up paths
        df["path"] = df["path"].apply(
            lambda x: str(Path(str(x).strip().strip('"').strip("'")))
        )
        df["path"] = df["path"].apply(lambda x: x if x.endswith("/") else x + "/")

        save_path = dataframe_path.with_stem(dataframe_path.stem + "_input")

    elif folder_path:
        folder_path = Path(str(folder_path).strip().strip('"').strip("'"))
        if not folder_path.is_dir():
            raise ValueError(f"{folder_path} is not a valid directory.")

        all_files = sorted([f.name for f in folder_path.glob("*") if f.is_file()])
        df = pd.DataFrame(
            {
                "file name": all_files,
                "path": [str(folder_path.resolve().as_posix()) + "/"] * len(all_files),
            }
        )

        # Save to parent of folder with timestamp
        today = datetime.now().strftime("%b_%d_%y")  # e.g., Jun_19_25
        save_path = folder_path.parent / f"input_df_{today}.xlsx"

    else:
        raise ValueError("You must provide either 'dataframe_path' or 'folder_path'.")

    # Add angular unit
    df["Angular unit"] = angular_unit

    # Save
    df.to_excel(save_path, index=False)
    print(f"Saved cleaned dataframe to: {save_path}")

    return df



# -*- coding: utf-8 -*-
"""SASM

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/13DPO65WEfQc3W86B_RrFrx8x81bafHSx
"""

from __future__ import absolute_import, division, print_function, unicode_literals

"""
Created on Jul 5, 2010

@author: Soren S. Nielsen

#******************************************************************************
# This file is part of RAW.
#
#    RAW is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    RAW is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with RAW.  If not, see <http://www.gnu.org/licenses/>.
#
#******************************************************************************
"""

from builtins import object, range, map, zip
# from io import open

import copy
# import os

import numpy as np
# from scipy import integrate as integrate

# raw_path = os.path.abspath(os.path.join('.', __file__, '..', '..'))
# if raw_path not in os.sys.path:
#     os.sys.path.append(raw_path)

# from .rawutils
# from src.rawutils.sasexceptions import sasexceptions as SASExceptions
from . import sasexceptions as SASExceptions


# import bioxtasraw.SASExceptions as SASExceptions


class FakeIntegrate:
    """
    Seems that scipy.integrate.trapz not a thing anymore so making a fake class to
    overwrite this without needing to change the code

    """

    trapz = staticmethod(np.trapz)


integrate = FakeIntegrate()


class SASM(object):
    """
    Small Angle Scattering Measurement (SASM) Object. Essentially a
    scattering profile with q, i, and uncertainty, plus a lot of metadata.

    Attributes
    ----------
    q: numpy.array
        The scaled q vector, without the trimming specified by
        :func:`setQrange`.
    i: numpy.array
        The scaled intensity vector, without the trimming specified by
        :func:`setQrange`.
    err: numpy.array
        The scaled error vector, without the trimming specified by
        :func:`setQrange`.
    q_err: numpy.array
        The scaled q error vector, without the trimming specified by
        :func:`setQrange`. Typically only used with SANS data.
    """

    def __init__(self, i, q, err, parameters, q_err=None):
        """
        Constructor

        Parameters
        ----------
        i: numpy.array
            The intensity vector.
        q: numpy.array
            The q vector.
        err: numpy.array
            The error vector.
        parameters: dict
            A dictionary of metadata for the object. This should contain at
            least {'filename': filename_with_no_path}. Other reserved keys are:
            'counters' : [(countername, value),...] Info from counter files
            'fileHeader' : [(label, value),...] Info from the header in the
            loaded file
        """

        # Raw intensity variables
        self._i_raw = np.array(i)
        self._q_raw = np.array(q)
        self._err_raw = np.array(err)
        self._parameters = parameters

        # Make an entry for analysis parameters i.e. Rg, I(0) etc:
        if "analysis" not in self._parameters:
            self._parameters["analysis"] = {}
        if "history" not in self._parameters:
            self._parameters["history"] = {}

        # Modified intensity variables
        self.i = self._i_raw.copy()
        self.q = self._q_raw.copy()
        self.err = self._err_raw.copy()

        # For SANS data with a qerr column
        # ** Could possibly delete this portion
        try:
            if q_err is not None:
                self._q_err_raw = np.array(q_err)
                self.q_err = self._q_err_raw.copy()
            else:
                self._q_err_raw = None
                self.q_err = None
        except Exception:
            self._q_err_raw = None
            self.q_err = None

        self._scale_factor = 1
        self._offset_value = 0
        self._q_scale_factor = 1

        # variables used for plot management
        self.item_panel = None
        self.plot_panel = None
        self.line = None
        self.err_line = None
        self.axes = None
        self.is_plotted = False
        self._selected_q_range = (0, len(self._q_raw))

        # Calculated values
        try:
            if len(self.q) > 0:
                self.total_intensity = integrate.trapz(self.getI(), self.getQ())
                self.mean_intensity = self.getI().mean()
            else:
                self.total_intensity = -1
                self.mean_intensity = -1

        except Exception as e:
            print(e)
            self.total_intensity = -1
            self.mean_intensity = -1

    def __deepcopy__(self, memo):
        # Raw intensity variables
        i_raw = copy.deepcopy(self._i_raw, memo)
        q_raw = copy.deepcopy(self._q_raw, memo)
        err_raw = copy.deepcopy(self._err_raw, memo)
        parameters = copy.deepcopy(self._parameters, memo)

        newsasm = SASM(i_raw, q_raw, err_raw, parameters)

        newsasm.setQrange(copy.deepcopy(self.getQrange(), memo))
        newsasm.setRawQErr(copy.deepcopy(self._q_err_raw, memo))

        newsasm.scale(copy.deepcopy(self.getScale(), memo))
        newsasm.offset(copy.deepcopy(self.getOffset(), memo))
        newsasm._q_scale_factor = copy.deepcopy(self._q_scale_factor, memo)

        newsasm._update()

        return newsasm

    def _update(self):
        """updates modified intensity after scale, normalization and offset changes"""

        self.i = (self._i_raw * self._scale_factor) + self._offset_value
        self.err = self._err_raw * abs(self._scale_factor)
        self.q = self._q_raw * self._q_scale_factor

        if self._q_err_raw is not None:
            self.q_err = self._q_err_raw * self._q_scale_factor

        # print self.err_line

        if self.err_line is not None:
            # Update errorbar positions
            caplines = self.err_line[0]
            barlinecols = self.err_line[1]

            yerr = self.err
            x = self.q
            y = self.i

            # Find the ending points of the errorbars
            error_positions = (x, y - yerr), (x, y + yerr)

            # Update the caplines
            for i, pos in enumerate(error_positions):
                caplines[i].set_data(pos)

            # Update the error bars
            barlinecols[0].set_segments(
                list(zip(list(zip(x, y - yerr)), list(zip(x, y + yerr))))
            )

        # Calculated values
        try:
            if len(self.q) > 0:
                self.total_intensity = integrate.trapz(self.getI(), self.getQ())
                self.mean_intensity = self.getI().mean()
            else:
                self.total_intensity = -1
                self.mean_intensity = -1

        except Exception as e:
            print(e)
            self.total_intensity = -1
            self.mean_intensity = -1

    def getScale(self):
        """
        Returns the scale factor for the profile.

        Returns
        -------
        scale: float
            The scale factor.
        """
        return self._scale_factor

    def getOffset(self):
        """
        Returns the offset for the profile.

        Returns
        -------
        offset: float
            The offset.
        """
        return self._offset_value

    def getLine(self):
        """
        Returns the plotted line for the profile. Only used in the RAW GUI.

        Returns
        -------
        line: matplotlib.lines.Line2D
            The plotted line.
        """
        return self.line

    def scaleRelative(self, relscale):
        """
        Applies a relative scale to the profile intensity. If the scale factor
        is currently 1, then this is the same as :func:`scale`. Otherwise,
        this scales relative to the current scale factor. For example, suppose
        the scale factor is currently 2. If a relative scale of 2 is provided,
        the resulting scale factor if 4. Scale factors are always positive.

        Parameters
        ----------
        relscale: float
            The relative scale factor to be applied to the the profile
            intensity and uncertainty.
        """
        self._scale_factor = abs(self._scale_factor * relscale)
        self._update()

    def scale(self, scale_factor):
        """
        Applies an absolute scale to the profile intensity. The scale factor
        supersedes the existing scale factor. For example, suppose the scale
        factor is currently 2. If a scale of 4 is provided, the resulting
        scale factor is 4 (not 8). Scale factors are always positive.

        Parameters
        ----------
        scale_factor: float
            The scale factor to be applied to the the profile intensity and
            uncertainty.
        """

        self._scale_factor = abs(scale_factor)
        self._update()

    def offset(self, offset_value):
        """
        Applies an absolute offset to the profile intensity. For example, if
        the offset is 1, then all the the intensities in the profile are
        increased by 1. Offset supersedes the existing offset, so if the
        current offset is 1, and an offset_value of 2 is provided, the
        resulting offset is 2 (not 3).

        Parameters
        ----------
        offset_value: float
            The offset to be applied to the profile intensity.
        """

        self._offset_value = offset_value
        self._update()

    def scaleRawQ(self, scale_factor):
        """
        Scales the raw q values. These are the q values without any scale
        applied. The raw q scale factor is not tracked, so this cannot easily
        be undone, unlike the :func:scaleQ function.

        Parameters
        ----------
        scale_factor: float
            The scale factor to be applied to the raw profile q values.
        """
        self._q_raw = self._q_raw * scale_factor
        self._update()

    def scaleQ(self, q_scale_factor):
        """
        Scales the profile q values by a factor. The scale factor
        supersedes the existing scale factor. For example, suppose the scale
        factor is currently 2. If a scale of 4 is provided, the resulting
        scale factor if 4 (not 8). Useful for converting between 1/A and 1/nm,
        for example.

        Parameters
        ----------
        q_scale_factor: float
            The scale factor to be applied to the profile q values.
        """

        self._q_scale_factor = q_scale_factor
        self._update()

    def scaleRelativeQ(self, relscale):
        """
        Applies a relative scale to the profile q. If the scale factor
        is currently 1, then this is the same as :func:`scale`. Otherwise,
        this scales relative to the current scale factor. For example, suppose
        the scale factor is currently 2. If a relative scale of 2 is provided,
        the resulting scale factor if 4.

        Parameters
        ----------
        relscale: float
            The relative scale factor to be applied to the the profile
            intensity and uncertainty.
        """
        self._q_scale_factor = self._q_scale_factor * relscale
        self._update()

    def reset(self):
        """
        Removes scale and offset values from the intensity, uncertainty, and q.
        """

        self.i = self._i_raw.copy()
        self.q = self._q_raw.copy()
        self.err = self._err_raw.copy()

        self._scale_factor = 1
        self._offset_value = 0
        self._q_scale_factor = 1

    def setQrange(self, qrange):
        """
        Sets the q range used for the profile. Useful for trimming leading or
        trailing values of the q profile that are not useful data.

        Parameters
        ----------
        qrange: tuple or list
            A tuple or list with two items. The first item is the starting
            index of the q vector to be used, the second item is the ending
            index of the q vector to be used, such that q[start:end] returns
            the desired q range.
        """
        if qrange[0] < 0 or qrange[1] > (len(self._q_raw)):
            msg = (
                "Qrange: " + str(qrange) + " is not a valid q-range for a "
                "q-vector of length " + str(len(self._q_raw) - 1)
            )
            raise SASExceptions.InvalidQrange(msg)
        else:
            self._selected_q_range = list(map(int, qrange))

            try:
                if len(self.q) > 0:
                    self.total_intensity = integrate.trapz(self.getI(), self.getQ())
                    self.mean_intensity = self.getI().mean()
                else:
                    self.total_intensity = -1
                    self.mean_intensity = -1

            except Exception as e:
                print(e)
                self.total_intensity = -1
                self.mean_intensity = -1

    def getQrange(self):
        """
        Returns the currently selected q range as described in
        :func:`setQrange`.

        Returns
        -------
        q_range: tuple
            A tuple with 2 indices, the start and end of the selected
            q range, such that q[start:end] returns the desired q range.
        """
        return self._selected_q_range

    def setAllParameters(self, new_parameters):
        """
        Sets the parameters dictionary, which contains the profile metadata,
        to the new input value.

        Parameters
        ----------
        new_parameters: dict
            A dictionary containing the new parameters.
        """
        self._parameters = new_parameters

    def getAllParameters(self):
        """
        Returns all of the metadata parameters associated with the profile as
        a dictionary.

        Returns
        -------
        parameters: dict
            The metadata associated with the profile.
        """
        return self._parameters

    def getParameter(self, key):
        """
        Gets a particular metadata parameter based on the provided key.

        Parameters
        ----------
        key: str
            A string that is a key in the parameters metadata dictionary.

        Returns
        -------
        parameter
            The parameter associated with the specified key. If the key is not
            in the parameter dictionary, None is returned.
        """

        if key in self._parameters:
            return self._parameters[key]
        else:
            return None

    def setParameter(self, key, value):
        """
        Sets a particular metadata parameter based on the provided key and value.

        Parameters
        ----------
        key: str
            The name of the new bit of metadata.
        value: object
            The value of the new bit of metadata. Could be anything that is
            an acceptable value for a dictionary.
        """
        self._parameters[key] = value

    def removeParameter(self, key):
        """
        Removes a particular metadata parameter based on the provided key.

        Parameters
        ----------
        key: str
            A string that is a key in the parameters metadata dictionary.
        """
        del self._parameters[key]

    def removeZingers(self, start_idx=0, window_length=10, stds=4.0):
        """
        Removes spikes (zingers) from radially averaged data based on smoothing
        of the intensity profile.

        Parameters
        ----------
        start_idx: int
            The initial index in the intensity to start the dezingering process.
        window_length: int
            The size of the window used to search for an replace spikes, as the
            number of intensity points.
        stds: The standard deviation threshold used to detect spikes.
        """

        intensity = self._i_raw

        for i in range(window_length + start_idx, len(intensity)):
            averaging_window = intensity[i - window_length : i]
            averaging_window_std = np.std(averaging_window)
            averging_window_mean = np.mean(averaging_window)

            threshold = averging_window_mean + (stds * averaging_window_std)

            if intensity[i] > threshold:
                intensity[i] = averging_window_mean

        self._update()

    def getRawQ(self):
        """
        Gets the raw q vector, without scaling based on the :func:`scaleQ` and
        without trimming based on :func:`setQrange`.

        Returns
        -------
        q_raw: numpy.array
            The raw q vector.
        """
        return self._q_raw

    def getRawI(self):
        """
        Gets the raw intensity vector, without scaling or offset from
        :func:`scale` and :func:`offset` and without trimming based on
        :func:`setQrange`.

        Returns
        -------
        i_raw: numpy.array
            The raw intensity vector.
        """
        return self._i_raw

    def getRawErr(self):
        """
        Gets the raw error vector, without scaling or offset from
        :func:`scale` and :func:`offset` and without trimming based on
        :func:`setQrange`.

        Returns
        -------
        err_raw: numpy.array
            The raw error vector.
        """
        return self._err_raw

    def getRawQErr(self):
        """
        Gets the raw q error vector, without scaling or offset from
        :func:`scale` and :func:`offset` and without trimming based on
        :func:`setQrange`.

        Returns
        -------
        q_err_raw: numpy.array
            The raw error vector.
        """
        return self._q_err_raw

    def setRawI(self, new_raw_i):
        """
        Sets the raw q vector. Will overwrite whatever q vector is already
        in the object! Typically only used during calibration.

        Parameters
        ----------
        new_raw_i: numpy.array
            The new intensity vector.
        """
        self._i_raw = new_raw_i

    def setRawQ(self, new_raw_q):
        """
        Sets the raw intensity vector. Will overwrite whatever intensity vector
        is already in the object! Typically only used during calibration.

        Parameters
        ----------
        new_raw_q: numpy.array
            The new q vector.
        """
        self._q_raw = new_raw_q

    def setRawErr(self, new_raw_err):
        """
        Sets the raw error vector. Will overwrite whatever error vector
        is already in the object! Typically only used during calibration.

        Parameters
        ----------
        new_raw_err: numpy.array
            The new error vector.
        """
        self._err_raw = new_raw_err

    def setRawQErr(self, new_raw_q_err):
        """
        Sets the raw q error vector. Will overwrite whatever error vector
        is already in the object! Typically only used during calibration.
        Q errors are typically only found in SANS data, and currently are
        only carried and written out with the data set, they are not used
        in any processing.

        Parameters
        ----------
        new_raw_err: numpy.array
            The new error vector.
        """
        self._q_err_raw = new_raw_q_err

    def setScaleValues(self, scale_factor, offset_value, q_scale_factor):
        """
        A convenience method that lets you set the scale offset, and q scale
        values all at once.

        Parameters
        ----------
        scale_factor: float
            The scale factor to be applied to the the profile intensity and
            uncertainty.
        offset_value: float
            The offset to be applied to the profile intensity.
        q_scale_factor: float
            The scale factor to be applied to the profile q values.
        """
        self._scale_factor = scale_factor
        self._offset_value = offset_value
        self._q_scale_factor = q_scale_factor
        self._update()

    def scaleRawIntensity(self, scale):
        """
        Scales the raw intensity and error values. These are the intensity and
        error values without any scale applied. The raw scale factor is not
        tracked, so this cannot easily be undone, unlike the :func:scaleQ function.

        Parameters
        ----------
        scale: float
            The scale factor to be applied to the raw profile intensity and
            error values.
        """
        self._i_raw = self._i_raw * scale
        self._err_raw = self._err_raw * scale
        self._update()

    def offsetRawIntensity(self, offset):
        """
        Offsets the raw intensity and error values. These are the intensity and
        error values without any offset applied. The raw offset factor is not
        tracked, so this cannot easily be undone, unlike the :func:scaleQ function.

        Parameters
        ----------
        offset: float
            The offset to be applied to the raw profile intensity and error
            values.
        """
        self._i_raw = self._i_raw + offset
        self._err_raw = self._err_raw
        self._update()

    def extractAll(self):
        """
        Extracts the raw and scaled q, intensity, and error, the scale and
        offset values, the selected q range, and the parameters in a single
        dictionary.

        Returns
        -------
        all_data: dict
            A dictionary with keys q_raw, i_raw, err_raw, q, i, err, scale_factor,
            offset_value, q_scale_factor, selected_qrange, and parameters, which
            correspond to those values from the SASM.
        """

        all_data = {}

        all_data["i_raw"] = self._i_raw
        all_data["q_raw"] = self._q_raw
        all_data["err_raw"] = self._err_raw
        all_data["q_err_raw"] = self._q_err_raw

        all_data["i"] = self.getI()
        all_data["q"] = self.getQ()
        all_data["err"] = self.getErr()
        all_data["q_err"] = self.getQErr()

        all_data["scale_factor"] = self._scale_factor
        all_data["offset_value"] = self._offset_value
        all_data["q_scale_factor"] = self._q_scale_factor

        all_data["selected_qrange"] = self._selected_q_range

        all_data["parameters"] = self._parameters

        return all_data

    def copy(self):
        """
        Creates a copy of the SASM, without scale information, using, for
        example, the scaled, trimmed intensity as the raw intensity for
        the new SASM.

        The preferred method of copying a profile is to use copy.deepcopy on
        the profile.

        Returns
        -------
        profile: bioxtasraw.SASM.SASM
            The copied profile
        """

        sasm = SASM(
            copy.deepcopy(self.i),
            copy.deepcopy(self.q),
            copy.deepcopy(self.err),
            copy.deepcopy(self._parameters),
        )
        sasm.setRawQErr(self._q_err_raw)

        return sasm

    def getMeanI(self):
        """
        Gets the mean intensity of the intensity vector.

        Returns
        -------
        mean_intensity: float
            The mean intensity.
        """
        return self.mean_intensity

    def getTotalI(self):
        """
        Gets the total integrated intensity of the intensity vector.

        Returns
        -------
        total_intensity: float
            The total intensity.
        """
        return self.total_intensity

    def getIofQ(self, qref):
        """
        Gets the intensity at a specific q value (or the closest such value in
        the q vector).

        Parameters
        ----------
        qref: float
            The reference q to get the intensity at.

        Returns
        -------
        intensity: float
            The intensity at the q point nearest the provided qref.
        """
        q = self.getQ()
        index = self.closest(q, qref)
        i = self.getI()
        intensity = i[index]

        return intensity

    def getIofQRange(self, q1, q2):
        """
        Gets the total integrated intensity in the q range from q1 to q2 (or
        the closest such values in the q vector).

        Parameters
        ----------
        q1: float
            The starting q value in the q range
        q2: float
            The ending q value in the q range.

        Returns
        -------
        total_intensity: float
            The total intensity.
        """
        q = self.getQ()
        index1 = self.closest(q, q1)
        index2 = self.closest(q, q2)
        i = self.getI()

        return integrate.trapz(i[index1 : index2 + 1], q[index1 : index2 + 1])

    @staticmethod
    def closest(qlist, q):
        """
        A convenience function which returns the index of the nearest
        q point in qlist to the input q value.

        Parameters
        ----------
        qlist: np.array
            The q list to search in.
        q: float
            The q value to search for in the qlist.

        Returns
        -------
        index: int
            The index of the q value in qlist closest to the input q.
        """
        return np.argmin(np.absolute(qlist - q))

    def getQ(self):
        """
        Gets the scaled, offset, trimmed q vector. Usually this is what you
        want to use to the get the q vector.
        """
        return self.q[self._selected_q_range[0] : self._selected_q_range[1]]

    def getI(self):
        """
        Gets the scaled, offset, trimmed intensity vector. Usually this is what
        you want to use to the get the intensity vector.
        """
        return self.i[self._selected_q_range[0] : self._selected_q_range[1]]

    def getErr(self):
        """
        Gets the scaled, offset, trimmed error vector. Usually this is what you
        want to use to the get the error vector.
        """
        return self.err[self._selected_q_range[0] : self._selected_q_range[1]]

    def getQErr(self):
        """
        Gets the scaled, offset, trimmed q error vector. Usually this is what you
        want to use to the get the q error vector. Q errors are usually only
        available with SANS data. Returns None if no q error has been defined.
        """

        if self.q_err is not None:
            q_err = self.q_err[self._selected_q_range[0] : self._selected_q_range[1]]
        else:
            q_err = None

        return q_err


class IFTM(object):
    """
    Inverse Fourier transform measurement (IFTM) object. Contains the P(r), r
    and error vectors, as well as the original data, the fit of the P(r) to
    the data, and all associated metadata.

    Attributes
    ----------
    r: numpy.array
        The r vector of the P(r) function.
    p: numpy.array
        The values of the P(r) function.
    err: numpy.array
        The errors of the P(r) function.
    q_orig: numpy.array
        The q vector of the input data.
    i_orig: numpy.array
        The intensity vector of the input data.
    err_orig: numpy.array
        The error vector of the input data.
    i_fit: numpy.array
        The intensity vector of the fit to the input data.
    q_extrap: numpy.array
        The q vector of the input data extrapolated to q=0.
    i_extrap: numpy.array
        The intensity vector of the fit to the input data extrapolated to q=0.
    """

    def __init__(
        self,
        p,
        r,
        err,
        i_orig,
        q_orig,
        err_orig,
        i_fit,
        parameters,
        i_extrap=[],
        q_extrap=[],
    ):
        """
        Constructor

        Parameters
        ----------
        p: numpy.array
            The input P(r) values.
        r:  numpy.array
            The input r values for the P(r) function.
        err: numpy.array
            The input error values for the P(r) function.
        i_orig: numpy.array
            The intensity values of the data used to do the IFT.
        q_orig: numpy.array
            The q values of the data used to do the IFT.
        err_orig: numpy.array
            The error values of the data used to do the IFT.
        i_fit: numpy.array
            The intensity values of the fit of the P(r) function to the data.
        parameters: dict
            A dictionary of the metadata. Should ontain at least {'filename':
            filename_with_no_path}
        i_extrap: numpy.array, optional
            The intensity values of the fit of the P(r) function to the data
            extrapolated to q=0. If not provided, an empty array is used.
        q_extrap: numpy.array, optional
            The q values of the input data extrapolated to q=0. If not
            provided, an empty array is used.
        """

        # Raw intensity variables
        self._r_raw = np.array(r)
        self._p_raw = np.array(p)
        self._err_raw = np.array(err)
        self._i_orig_raw = np.array(i_orig)
        self._q_orig_raw = np.array(q_orig)
        self._err_orig_raw = np.array(err_orig)
        self._i_fit_raw = np.array(i_fit)
        self._i_extrap_raw = np.array(i_extrap)
        self._q_extrap_raw = np.array(q_extrap)
        self._parameters = parameters

        # Make an entry for analysis parameters i.e. Rg, I(0) etc:
        # if 'analysis' not in self._parameters:
        #     self._parameters['analysis'] = {}
        # if 'history' not in self._parameters:
        #     self._parameters['history'] = {}

        # Modified intensity variables
        self.r = self._r_raw.copy()
        self.p = self._p_raw.copy()
        self.err = self._err_raw.copy()

        self.i_orig = self._i_orig_raw.copy()
        self.q_orig = self._q_orig_raw.copy()
        self.err_orig = self._err_orig_raw.copy()

        self.i_fit = self._i_fit_raw.copy()

        self.i_extrap = self._i_extrap_raw.copy()
        self.q_extrap = self._q_extrap_raw.copy()

        # variables used for plot management
        self.item_panel = None

        self.plot_panel = None

        self.r_line = None
        self.qo_line = None
        self.qf_line = None

        self.r_origline = None
        self.qo_origline = None
        self.qf_origline = None

        self.r_err_line = None
        self.qo_err_line = None

        self.r_axes = None
        self.qo_axes = None
        self.qf_axes = None

        self.canvas = None

        self.is_plotted = False

    def __deepcopy__(self, memo):
        p = copy.deepcopy(self._p_raw, memo)
        r = copy.deepcopy(self._r_raw, memo)
        err = copy.deepcopy(self._err_raw, memo)

        i_orig = copy.deepcopy(self._i_orig_raw, memo)
        q_orig = copy.deepcopy(self._q_orig_raw, memo)
        err_orig = copy.deepcopy(self._err_orig_raw, memo)

        i_fit = copy.deepcopy(self._i_fit_raw, memo)

        i_extrap = copy.deepcopy(self._i_extrap_raw, memo)
        q_extrap = copy.deepcopy(self._q_extrap_raw, memo)

        parameters = copy.deepcopy(self._parameters, memo)

        new_iftm = IFTM(
            p, r, err, i_orig, q_orig, err_orig, i_fit, parameters, i_extrap, q_extrap
        )

        return new_iftm

    def getScale(self):
        """
        Returns the scale factor for the P(r) function.

        Returns
        -------
        scale: float
            The scale factor.
        """
        return self._scale_factor

    def getOffset(self):
        """
        Returns the offset for the P(r) function.

        Returns
        -------
        offset: float
            The offset.
        """
        return self._offset_value

    def getLine(self):
        """
        Returns the plotted line for the P(r) function. Only used in the RAW GUI.

        Returns
        -------
        line: matplotlib.lines.Line2D
            The plotted line.
        """
        return self.line

    def setAllParameters(self, new_parameters):
        """
        Sets the parameters dictionary, which contains the IFT metadata,
        to the new input value.

        Parameters
        ----------
        new_parameters: dict
            A dictionary containing the new parameters.
        """
        self._parameters = new_parameters

    def getAllParameters(self):
        """
        Returns all of the metadata parameters associated with the IFT as
        a dictionary.

        Returns
        -------
        parameters: dict
            The metadata associated with the IFT.
        """
        return self._parameters

    def getParameter(self, key):
        """
        Gets a particular metadata parameter based on the provided key.

        Parameters
        ----------
        key: str
            A string that is a key in the parameters metadata dictionary.

        Returns
        -------
        parameter
            The parameter associated with the specified key. If the key is not
            in the parameter dictionary, None is returned.
        """

        if key in self._parameters:
            return self._parameters[key]
        else:
            return None

    def setParameter(self, key, value):
        """
        Sets a particular metadata parameter based on the provided key and value.

        Parameters
        ----------
        key: str
            The name of the new bit of metadata.
        value: object
            The value of the new bit of metadata. Could be anything that is
            an acceptable value for a dictionary.
        """
        self._parameters[key] = value

    def extractAll(self):
        """
        Extracts the raw and scaled q, intensity, and error, the scale and
        offset values, the selected q range, and the parameters in a single
        dictionary.

        Returns
        -------
        all_data: dict
            A dictionary with keys r_raw, p_raw, err_raw, i_orig_raw, q_orig_raw,
            err_orig_raw, i_fit_raw, i_extrap_raw, q_extrap_raw, and parameters,
            which correspond to those values from the IFTM.
        """

        all_data = {}

        all_data["r_raw"] = self._r_raw
        all_data["p_raw"] = self._p_raw
        all_data["err_raw"] = self._err_raw

        all_data["i_orig_raw"] = self._i_orig_raw
        all_data["q_orig_raw"] = self._q_orig_raw
        all_data["err_orig_raw"] = self._err_orig_raw

        all_data["i_fit_raw"] = self._i_fit_raw
        all_data["i_extrap_raw"] = self._i_extrap_raw
        all_data["q_extrap_raw"] = self._q_extrap_raw

        all_data["parameters"] = self._parameters

        return all_data

        pass

    def copy(self):
        """
        Creates a copy of the IFT.

        Returns
        -------
        ift: bioxtasraw.SASM.IFTM
            The copied IFTM
        """

        iftm_copy = IFTM(
            copy.deepcopy(self._p_raw),
            copy.deepcopy(self._r_raw),
            copy.deepcopy(self._err_raw),
            copy.deepcopy(self._i_orig_raw),
            copy.deepcopy(self._q_orig_raw),
            copy.deepcopy(self._err_orig_raw),
            copy.deepcopy(self._i_fit_raw),
            copy.deepcopy(self._parameters),
            copy.deepcopy(self._i_extrap_raw),
            copy.deepcopy(self._q_extrap_raw),
        )

        return iftm_copy


# Uncommented 03/29
def postProcessSasm(sasm, raw_settings):
    pass

    # raw_settings = False #added bc not using settings module

    # if raw_settings.get('ZingerRemoval'):
    #     std = raw_settings.get('ZingerRemoveSTD')
    #     winlen = raw_settings.get('ZingerRemoveWinLen')
    #     start_idx = raw_settings.get('ZingerRemoveIdx')

    #     sasm.removeZingers(start_idx, winlen, std)



"""This contains functions relating to the use of Machine Learning in SAXS data analysis.
Including the clustering and the prediction of Dmax using a MLP regressor
trained on SASBDB data
"""

import numpy as np
import pandas as pd
import logging
from joblib import load
from scipy.integrate import quad, trapezoid
from saxs_assistant import models
from importlib import resources


def load_model(filename: str):
    with resources.path(models, filename) as path:
        bundle = load(path)
    return bundle


def compute_franke_features(q, I, rg, i0, sample_id, df, j):
    """
    Computes Franke shape features and stores them only in the dataframe.
    Does NOT touch plot_data.
    """
    try:
        V = get_franke_features2(q, rg, I, i0)
        feature_labels = ["V_1.7", "V_3", "V_4", "V_5", "V_6", "V_7", "V_8"]
        for k, v in zip(feature_labels, V):
            df.loc[df.index[j], k] = v

    except Exception as e:
        logging.warning(f"Franke feature calculation failed for {sample_id}: {e}")

    return df


def predict_dmax_from_features_only(
    df, j, bundle_path="dmax_predictor_2bundle_05262025.joblib"
):
    # bundle_path=r"models\dmax_predictor_2bundle_05262025.joblib" no longer using cause resources
    """
    Loads model bundle and predicts Dmax for row j of df.
    Returns the predicted Dmax and feature values used.
    The name of the model bundle must remain the same for it to work
    """
    try:
        # bundle = load(bundle_path)
        bundle = load_model(bundle_path)
        model = bundle["model"]
        scaler = bundle["scaler"]
        feature_names = bundle["features"]

        # Mapping of model-trained features to your df column names
        feature_map = {
            "guinier_rg": "Final Rg",
            "V3 OG": "V_3",
            "V4 OG": "V_4",
            "V5 OG": "V_5",
            "min_qrg OG": "Final qRg min",
        }

        # Extract and scale feature vector
        feature_vector = np.array(
            [df.loc[df.index[j], feature_map[feat]] for feat in feature_names]
        ).reshape(1, -1)

        # feature_vector should be a 1D or 2D array (e.g., shape (1, n_features))
        # feature_names should be the exact names used during training
        scaled = scaler.transform(pd.DataFrame(feature_vector, columns=feature_names))

        prediction = model.predict(scaled)[0]

        # Return result + raw features used
        return prediction, {
            feat: df.loc[df.index[j], feature_map[feat]] for feat in feature_names
        }

    except Exception as e:
        logging.warning(f"Dmax prediction failed for index {j}: {e}")
        return np.nan, {"error": str(e)}


def assign_gmm_clusters(df, j, bundle_path="gmm_cluster_2bundle_05262025.joblib"):
    # bundle_path=r"models\gmm_cluster_2bundle_05262025.joblib" no longer using cause resources
    """
    Loads model bundle and predicts cluster probabilities using GMM for row j of df.
    Returns the predicted cluster and feature values used.
    The name of the model bundle must remain the same for it to work
    """

    try:
        # bundle = load(bundle_path)
        bundle = load_model(bundle_path)
        model = bundle["model"]
        scaler = bundle["scaler"]
        features = bundle["features"]

        # Map feature names
        feature_map = {"V3 OG": "V_3", "V4 OG": "V_4", "V5 OG": "V_5"}

        raw_feature_values = [df.loc[df.index[j], feature_map[ft]] for ft in features]
        scaled = scaler.transform([raw_feature_values])

        cluster = int(model.predict(scaled)[0])
        probs = model.predict_proba(scaled)[0]

        df.loc[df.index[j], "GMM Cluster"] = cluster
        for i, p in enumerate(probs):
            df.loc[df.index[j], f"Cluster Prob {i}"] = p

        return {
            "Cluster": cluster,
            "Probabilities": {f"Cluster {i}": float(p) for i, p in enumerate(probs)},
        }

    except Exception as e:
        logging.warning(f"GMM clustering failed for index {j}: {e}")
        return None


def get_franke_features2(q, rg, I_q, i0):
    """
    Getting features used for shape classification by Franke in Machine learning methods
    for X-ray scattering data analysis from biomacromolecular solutions
    """
    k_y = ((q * rg) ** 2) * I_q / i0
    k_x = q * rg
    # Find indx of qRg = 3, 4, 5
    ind_1_7, ind_3, ind_4, ind_5, ind_6, ind_7, ind_8 = (
        np.argmin(abs(k_x - 1.73)),
        np.argmin(abs(k_x - 3)),
        np.argmin(abs(k_x - 4)),
        np.argmin(abs(k_x - 5)),
        np.argmin(abs(k_x - 6)),
        np.argmin(abs(k_x - 7)),
        np.argmin(abs(k_x - 8)),
    )
    k_y_1_7, k_y_3, k_y_4, k_y_5, k_y_6, k_y_7, k_y_8 = (
        k_y[:ind_1_7],
        k_y[:ind_3],
        k_y[:ind_4],
        k_y[:ind_5],
        k_y[:ind_6],
        k_y[:ind_7],
        k_y[:ind_8],
    )
    k_x_1_7, k_x_3, k_x_4, k_x_5, k_x_6, k_x_7, k_x_8 = (
        k_x[:ind_1_7],
        k_x[:ind_3],
        k_x[:ind_4],
        k_x[:ind_5],
        k_x[:ind_6],
        k_x[:ind_7],
        k_x[:ind_8],
    )
    # find Q' (normalized porod invariant)
    Q_1_7, Q_3, Q_4, Q_5, Q_6, Q_7, Q_8 = (
        trapezoid(k_y_1_7, k_x_1_7),
        trapezoid(k_y_3, k_x_3),
        trapezoid(k_y_4, k_x_4),
        trapezoid(k_y_5, k_x_5),
        trapezoid(k_y_6, k_x_6),
        trapezoid(k_y_7, k_x_7),
        trapezoid(k_y_8, k_x_8),
    )
    V_1_7, V_3, V_4, V_5, V_6, V_7, V_8 = (
        (2 * np.pi**2) / Q_1_7,
        (2 * np.pi**2) / Q_3,
        (2 * np.pi**2) / Q_4,
        (2 * np.pi**2) / Q_5,
        (2 * np.pi**2) / Q_6,
        (2 * np.pi**2) / Q_7,
        (2 * np.pi**2) / Q_8,
    )

    return V_1_7, V_3, V_4, V_5, V_6, V_7, V_8



import os
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib import gridspec
from matplotlib.ticker import FormatStrFormatter
import numpy as np
from ..features import min_max_function
from ..rg_tools import get_residuals_guinier_plot
from joblib import load


def plot_solved_summary(
    plot_data_path, output_folder="./summary_plots", pdf_name="solved_summary.pdf"
):
    # Load data internally
    if isinstance(plot_data_path, str):
        plot_data = load(plot_data_path)
    else:
        raise ValueError("plot_data_path should be a string path to a .joblib file")

    os.makedirs(output_folder, exist_ok=True)
    pdf_path = os.path.join(output_folder, pdf_name)

    with PdfPages(pdf_path) as pdf:
        for sample_id, plot_item in plot_data.items():
            if "Flagged" in plot_item:
                continue  # Skip flagged entries

            file_name = plot_item["profile"]["title"].split(" ")[-1]
            sample_short = file_name.split(".")[0]

            # --- Setup master figure for one sample (4 stacked rows) ---
            fig = plt.figure(figsize=(8.5, 12))
            outer_gs = gridspec.GridSpec(
                4, 1, hspace=0.4, height_ratios=[1, 1, 1.1, 1.1]
            )

            # --- Row 1: Profile, GPA, Dimless Kratky ---
            gs1 = gridspec.GridSpecFromSubplotSpec(
                1, 3, subplot_spec=outer_gs[0], wspace=0.4
            )
            axs = [fig.add_subplot(gs1[0, i]) for i in range(3)]

            axs[0].errorbar(
                plot_item["profile"]["x"],
                plot_item["profile"]["y"],
                plot_item["profile"]["yerr"],
                ecolor="g",
                fmt="bo",
                markersize=2,
                elinewidth=0.5,
            )
            axs[0].set_yscale("log")
            axs[0].set_title(file_name, fontsize=7.5)
            axs[0].set_xlabel(r"$q$ $(\mathrm{\AA}^{-1})$", fontsize=7)
            axs[0].set_ylabel(r"$log(I(q))$", fontsize=7)
            axs[0].tick_params(axis="both", which="major", labelsize=7)

            axs[1].plot(
                plot_item["GPA"]["x"], plot_item["GPA"]["y"], "ob", markersize=2
            )
            axs[1].set_title("GPA", fontsize=7.5)
            axs[1].set_xlabel(r"$q$ $(\mathrm{\AA}^{-2})$", fontsize=7)
            axs[1].xaxis.get_offset_text().set_fontsize(6.5)
            axs[1].set_ylabel(f"${plot_item['GPA']['ylabel']}$", fontsize=7)
            axs[1].tick_params(axis="both", which="major", labelsize=7)

            axs[2].plot(
                plot_item["Dimless Kratky"]["x"],
                plot_item["Dimless Kratky"]["y"],
                "ob",
                markersize=2,
            )
            axs[2].set_title("Dimensionless Kratky", fontsize=7.5)
            axs[2].set_xlabel(r"$qR_g$", fontsize=7)
            axs[2].set_ylabel(r"$(qR_g)^2 I(q)/I(0)$", fontsize=7)
            axs[2].axvline(x=3**0.5, color="red", linestyle="--", linewidth=1)
            axs[2].axhline(
                y=1.1, color="red", linestyle="--", label="Globular Crosshairs"
            )
            axs[2].legend(fontsize=6.5, loc="best")
            axs[2].set_xlim([0, 6])
            axs[2].set_ylim([0, 2])
            axs[2].tick_params(axis="both", which="major", labelsize=7)

            # --- Row 2: GPA fits (Auto and Method 1) and GMM Clustering ---
            gs2 = gridspec.GridSpecFromSubplotSpec(
                1, 3, subplot_spec=outer_gs[1], width_ratios=[2.8, 2.8, 0.9], wspace=0.4
            )

            for idx, label, rg_key in zip(
                [0, 1],
                ["Auto $R_g$", "PDDF-Informed"],
                ["Auto Rg", "Rg Method 1 Final"],
            ):
                ax = fig.add_subplot(gs2[0, idx])
                rg = plot_item[rg_key]["Rg"] if idx == 0 else plot_item[rg_key]["Rg"][0]
                i0 = plot_item[rg_key]["i0"] if idx == 0 else plot_item[rg_key]["i0"][0]
                nmin = (
                    plot_item[rg_key]["nmin"]
                    if idx == 0
                    else plot_item[rg_key]["nmin"][0]
                )
                nmax = (
                    plot_item[rg_key]["nmax"]
                    if idx == 0
                    else plot_item[rg_key]["nmax"][0]
                )
                x_maxs, y_maxs, plot_params = get_dim_GPA2(
                    plot_item["profile"]["y"],
                    plot_item["profile"]["x"],
                    rg,
                    i0,
                    nmin,
                    nmax,
                )
                gpa_ = plot_params["data"]
                ax.plot(gpa_["x"], gpa_["y"], "bo", markersize=3, label="Data")
                ax.plot(
                    gpa_["included_x"],
                    gpa_["included_y"],
                    "o",
                    color="orange",
                    markersize=3,
                    label="Included Data",
                )
                ax.plot(
                    gpa_["guinier_x"], gpa_["guinier_y"], "g--", label="Calculated I(q)"
                )
                for x_val, y_data, y_model in zip(
                    gpa_["x"], gpa_["y"], gpa_["guinier_y"]
                ):
                    ax.plot(
                        [x_val, x_val],
                        [y_data, y_model],
                        color="red",
                        linewidth=1,
                        alpha=0.7,
                    )
                ax.axvline(x=1.5, linestyle="--", color="purple")
                ax.axhline(
                    y=np.log(0.7428),
                    linestyle="--",
                    color="purple",
                    label="Ideal Peak Location",
                )
                ax.set_xlabel(r"$(q R_g)^2$", fontsize=7)
                ax.set_ylabel("ln[$q$$R_g$$I(q)$/$I(0)$]", fontsize=7)
                ax.set_xlim([0, 4])
                ax.set_ylim([-2, 0])
                ax.set_title(label, fontsize=7.5)
                ax.legend(fontsize=6.5)
                ax.tick_params(labelsize=7)
                if idx == 1:
                    for spine in ax.spines.values():
                        spine.set_edgecolor("green")
                        spine.set_linewidth(2)

            ax3 = fig.add_subplot(gs2[0, 2])
            prob_found = plot_item["GMM Clustering"]["Probabilities"]
            assigned_cluster = plot_item["GMM Clustering"]["Cluster"]
            probs = [prob_found[f"Cluster {i}"] for i in range(5)]
            labels = ["0 (Fd)", "1 (Pd)", "2 (Ds)", "3 (Gb)", "4 (Fx)"]
            colors = ["red", "skyblue", "green", "orange", "purple"]
            ax3.barh(range(len(probs)), probs, color=colors)
            ax3.set_xlim(0, 1)
            ax3.set_yticks([])
            ax3.set_xticks([0, 0.5, 1.0])
            ax3.set_title("Cluster\nProbs", fontsize=7.5)
            ax3.tick_params(axis="x", labelsize=7)
            for i, (label, prob) in enumerate(zip(labels, probs)):
                fontweight = "bold" if i == assigned_cluster else "normal"
                ax3.text(
                    -0.02,
                    i,
                    label,
                    va="center",
                    ha="right",
                    fontsize=6,
                    fontweight=fontweight,
                    transform=ax3.get_yaxis_transform(),
                )
                ax3.text(
                    prob + 0.02,
                    i,
                    f"{prob * 100:.1f}%",
                    va="center",
                    ha="left",
                    fontsize=7,
                )
            for spine in ["top", "right"]:
                ax3.spines[spine].set_visible(False)

            # --- Row 3: Auto & PDDF-Informed Guinier Residuals ---
            gs3 = gridspec.GridSpecFromSubplotSpec(
                2,
                2,
                subplot_spec=outer_gs[2],
                height_ratios=[2.5, 1],
                wspace=0.25,
                hspace=0.1,
            )

            # ---------- Left Column: Auto Guinier ----------
            results_auto = get_residuals_guinier_plot(
                plot_item["profile"]["y"],
                plot_item["profile"]["x"],
                plot_item["profile"]["yerr"],
                plot_item["Auto Rg"]["nmin"],
                plot_item["Auto Rg"]["nmax"],
            )
            ax_top = fig.add_subplot(gs3[0, 0])
            ax_bot = fig.add_subplot(gs3[1, 0], sharex=ax_top)

            ax_top.plot(
                results_auto["x_all"],
                results_auto["y_all"],
                "o",
                color="gray",
                alpha=0.8,
                markersize=2,
            )
            ax_top.plot(
                results_auto["x_fit"],
                results_auto["y_fit"],
                "o",
                color="blue",
                markersize=2,
            )
            ax_top.plot(
                results_auto["x_fit"],
                results_auto["y_model"],
                "-",
                color="red",
                linewidth=1,
                label=r"$R^2$ " + str(round(plot_item["Auto Rg"]["R2"], 3)),
            )
            ax_top.set_title(
                "Auto Guinier: $R_g$: "
                + str(round(plot_item["Auto Rg"]["Rg"], 2))
                + r"$\pm$"
                + str(round(plot_item["Auto Rg"]["Rg Err"], 2)),
                fontsize=7,
            )

            ax_top.set_ylabel(r"$\ln[I(q)]$", fontsize=7)
            ax_top.tick_params(axis="both", which="major", labelsize=7)
            ax_top.axvline(
                x=results_auto["x_fit"][0], linestyle="--", color="red", linewidth=1
            )
            ax_top.axvline(
                x=results_auto["x_fit"][-1], linestyle="--", color="red", linewidth=1
            )
            ax_top.set_ylim(
                [results_auto["y_model"].min() - 0.5, results_auto["y_all"].max() + 0.5]
            )
            ax_top.set_xlim(
                [0, results_auto["x_fit"][-1] + (results_auto["x_fit"].ptp() * 0.10)]
            )
            ax_top.tick_params(labelbottom=False)  # hides x-tick labels on top
            ax_top.legend(fontsize=6.5, loc="upper right")

            ax_bot.plot(
                results_auto["x_fit"], results_auto["residuals"], "b-", linewidth=1
            )
            ax_bot.axhline(y=0, linestyle="--", color="red", linewidth=1)
            ax_bot.set_xlabel(r"$q^2$ $(\mathrm{\AA}^{-2})$", fontsize=7)
            ax_bot.xaxis.set_major_formatter(FormatStrFormatter("%.4f"))
            ax_bot.set_ylabel(r"$\Delta \ln[I(q)] / \sigma(q)$", fontsize=7)
            ax_bot.tick_params(axis="both", which="major", labelsize=7)

            # ---------- Right Column: Method 1 ----------
            results_m1 = get_residuals_guinier_plot(
                plot_item["profile"]["y"],
                plot_item["profile"]["x"],
                plot_item["profile"]["yerr"],
                plot_item["Rg Method 1 Final"]["nmin"][0],
                plot_item["Rg Method 1 Final"]["nmax"][0],
            )
            ax_top = fig.add_subplot(gs3[0, 1])
            ax_bot = fig.add_subplot(gs3[1, 1], sharex=ax_top)

            # ax_top, ax_bot = axes[0, 1], axes[1, 1]

            ax_top.plot(
                results_m1["x_all"],
                results_m1["y_all"],
                "o",
                color="gray",
                alpha=0.8,
                markersize=2,
            )
            ax_top.plot(
                results_m1["x_fit"],
                results_m1["y_fit"],
                "o",
                color="blue",
                markersize=2,
            )
            ax_top.plot(
                results_m1["x_fit"],
                results_m1["y_model"],
                "-",
                color="red",
                linewidth=1,
                label=r"$R^2$ "
                + str(round(plot_item["Rg Method 1 Final"]["fit_r2"][0], 3)),
            )
            ax_top.set_title(
                "PDDF-Informed: $R_g$: "
                + str(round(plot_item["Rg Method 1 Final"]["Rg"][0], 2))
                + r"$\pm$"
                + str(round(plot_item["Rg Method 1 Final"]["Rg Err"][0], 2)),
                fontsize=7,
            )

            ax_top.set_ylabel(r"$\ln[I(q)]$", fontsize=7)
            ax_top.tick_params(axis="both", which="major", labelsize=7)
            ax_top.axvline(
                x=results_m1["x_fit"][0], linestyle="--", color="red", linewidth=1
            )
            ax_top.axvline(
                x=results_m1["x_fit"][-1], linestyle="--", color="red", linewidth=1
            )
            ax_top.set_ylim(
                [results_m1["y_model"].min() - 0.5, results_m1["y_all"].max() + 0.5]
            )
            ax_top.set_xlim(
                [0, results_m1["x_fit"][-1] + (results_m1["x_fit"].ptp() * 0.10)]
            )
            ax_top.tick_params(labelbottom=False)  # hides x-tick labels on top
            ax_top.legend(fontsize=6.5, loc="upper right")

            ax_bot.plot(results_m1["x_fit"], results_m1["residuals"], "b-", linewidth=1)
            ax_bot.axhline(y=0, linestyle="--", color="red", linewidth=1)
            ax_bot.set_xlabel(r"$q^2$ $(\mathrm{\AA}^{-2})$", fontsize=7)
            ax_bot.xaxis.set_major_formatter(FormatStrFormatter("%.4f"))
            ax_bot.set_ylabel(r"$\Delta \ln[I(q)] / \sigma(q)$", fontsize=7)
            ax_bot.tick_params(axis="both", which="major", labelsize=7)

            # Row 4 PDDF

            # ======= Left panel: PDDF Fit + Residuals =======
            gs4 = gridspec.GridSpecFromSubplotSpec(
                1, 2, subplot_spec=outer_gs[3], width_ratios=[3.2, 2.8], wspace=0.3
            )
            gs4_left = gridspec.GridSpecFromSubplotSpec(
                2, 1, subplot_spec=gs4[0], height_ratios=[2, 1], hspace=0.05
            )

            # Top: Fit
            # ax_fit = fig.add_subplot(gs4_left[0])
            # [your plotting code for fit...]

            # ======= Right panel: PDDF Plot =======
            # ax_pddf = fig.add_subplot(gs4[1])
            # [your PDDF curve plotting code...]

            # Top: Fit
            ax_fit = fig.add_subplot(gs4_left[0])
            ax_fit.plot(
                plot_item["pr_plot"]["pr_q_orig"],
                plot_item["pr_plot"]["pr_i_orig"],
                "o",
                color="blue",
                alpha=0.8,
                markersize=2,
                label="Data",
                linewidth=1,
            )
            ax_fit.plot(
                plot_item["pr_residuals"]["q_extrapolated"][1:],
                plot_item["pr_residuals"]["i_fit"],
                color="red",
                linewidth=1,
                label="Fitted Region",
            )
            ax_fit.set_yscale("log")
            ax_fit.set_title("Data/Fit", fontsize=7.5)

            ax_fit.set_ylabel(r"$I(q)$", fontsize=7)
            ax_fit.tick_params(axis="both", which="major", labelsize=7)
            ax_fit.legend(fontsize=7)

            # Bottom: Residuals
            ax_res = fig.add_subplot(gs4_left[1], sharex=ax_fit)
            ax_res.plot(
                plot_item["pr_residuals"]["residuals_x"],
                plot_item["pr_residuals"]["residuals_y"],
                "b-",
                linewidth=1,
            )
            ax_res.axhline(y=0, linestyle="--", color="red", linewidth=1)

            ax_res.set_xlabel(r"$q$ $(\mathrm{\AA})$", fontsize=7)
            ax_res.set_ylabel(r"$\Delta [I(q)]/\sigma(q)$", fontsize=7)
            ax_res.tick_params(axis="both", which="major", labelsize=7)

            # ======= Right panel: PDDF Plot =======
            predicted_dmax = plot_item["Predicted Dmax"]["prediction"]
            BIFT_dmax = plot_item["pr_plot"]["x"][-1]
            diff_dmax = (abs(predicted_dmax - BIFT_dmax) / BIFT_dmax) * 100

            ax_pddf = fig.add_subplot(gs4[1])
            ax_pddf.plot(
                plot_item["pr_plot"]["x"],
                plot_item["pr_plot"]["y"],
                "bo-",
                markersize=2,
                linewidth=1.5,
                label=f"BIFT $D_{{max}}$ = {BIFT_dmax:.1f} Å",
            )

            ax_pddf.set_xlabel(r"$r$ $(\mathrm{\AA})$", fontsize=7)
            ax_pddf.set_ylabel(r"$P(r)$", fontsize=7)
            xmax = max(predicted_dmax, BIFT_dmax)
            ax_pddf.set_xlim(0, xmax * 1.05)
            ax_pddf.set_ylim(bottom=0)
            ax_pddf.tick_params(axis="both", which="major", labelsize=7.5)
            ax_pddf.set_title(
                "PDDF $R_g$:"
                + str(round(plot_item["pr_plot"]["Rg"], 2))
                + "+/-"
                + str(round(plot_item["pr_plot"]["Rg err"], 2)),
                fontsize=7.5,
            )

            # Optional: vertical Dmax line
            if diff_dmax < 20:
                ax_pddf.axvline(
                    predicted_dmax,
                    color="green",
                    linestyle="--",
                    linewidth=1.5,
                    alpha=0.7,
                    label=f"Predicted $D_{{max}}$ = {predicted_dmax:.1f} Å",
                )
            elif diff_dmax >= 20:
                ax_pddf.axvline(
                    predicted_dmax,
                    color="red",
                    linestyle="--",
                    linewidth=1.5,
                    alpha=0.7,
                    label=f"Predicted $D_{{max}}$ = {predicted_dmax:.1f} Å",
                )

            ax_pddf.legend(fontsize=7)

            # fig.savefig(f'SI1_row4_Pr_residuals_{id}.png', dpi=300, bbox_inches='tight')
            # plt.show()

            pdf.savefig(fig)
            plt.close(fig)

    print(f"Saved PDF summary to {pdf_path}")
    return pdf_path


def get_dim_GPA2(I, q, rg, i0, nmin, nmax, plot=False):
    """Another Dimensionless GPA function that uses the same logic as get_dim_GPA but with different outputs, in
    the format of dictionary for plotting.
    Args:
        I (array): Intensity data.
        q (array): q values.
        rg (float): Radius of gyration.
        i0 (float): Intensity at zero angle.
        nmin (int): Minimum index for the range of interest.
        nmax (int): Maximum index for the range of interest.
        Returns:
        x_maxs (float): Maximum x value.
        y_maxs (float): Maximum y value.
        plot_params (dict): Dictionary containing plot parameters.
    """
    y = np.log((q * rg * I) / i0)
    x = (q * rg) ** 2
    index_max_y = np.argmax(y)
    index_max_plot = np.argmin(abs(x - 4))
    x_mins, x_maxs, y_mins, y_maxs = min_max_function(
        y[:index_max_plot], x[:index_max_plot]
    )

    # Prepare plot data for saving
    plot_params = {
        "type": "Dimensionless GPA",
        "data": {
            "x": x,
            "y": y,
            "included_x": x[nmin : nmax + 1],
            "included_y": y[nmin : nmax + 1],
            "guinier_x": x,
            "guinier_y": np.log((q * rg * (i0 * np.exp(-((rg**2 * q**2) / 3)))) / i0),
        },
        "lines": [
            {
                "x": [1.5, 1.5],
                "y": [-2, 0],
                "style": "--",
                "label": "Ideal Vertical Line",
            },
            {
                "x": [0, 4],
                "y": [np.log(0.7428), np.log(0.7428)],
                "style": "--",
                "label": "Ideal Horizontal Line",
            },
        ],
        "labels": {
            "title": "Dim GPA Rg: Peak",
            "xlabel": "$(Rgq)^2$",
            "ylabel": "ln[qRgI(q)/I(0)]",
        },
        "limits": {"xlim": [0, 4], "ylim": [-2, 0]},
    }

    if plot == True:
        fig, ax = plt.subplots()
        # Plot data
        ax.plot(plot_params["data"]["x"], plot_params["data"]["y"], "o", label="Data")
        ax.plot(
            plot_params["data"]["included_x"],
            plot_params["data"]["included_y"],
            "o",
            label="Included",
        )
        ax.plot(
            plot_params["data"]["guinier_x"],
            plot_params["data"]["guinier_y"],
            "--",
            label="Guinier Approximation",
        )
        # Plot lines
        for line in plot_params["lines"]:
            ax.plot(line["x"], line["y"], line["style"], label=line["label"])
        # Set labels and limits
        ax.set_title(plot_params["labels"]["title"])
        ax.set_xlabel(plot_params["labels"]["xlabel"])
        ax.set_ylabel(plot_params["labels"]["ylabel"])
        ax.set_xlim(plot_params["limits"]["xlim"])
        ax.set_ylim(plot_params["limits"]["ylim"])
        ax.legend()
        plt.show()

    return x_maxs, y_maxs, plot_params
